# django settings for tests
SECRET_KEY = "test"
INSTALLED_APPS = (
    "django.contrib.contenttypes",
    "openstates.data",
    "openstates.reports",
)
DATABASES = {
    "default": {
        "ENGINE": "django.contrib.gis.db.backends.postgis",
        "NAME": "test",
        "USER": "test",
        "PASSWORD": "test",
        "HOST": "localhost",
    }
}
MIDDLEWARE_CLASSES = ()



from django.db import transaction
from django.core.management.base import BaseCommand

from ...divisions import Division as FileDivision
from ...models import Division


def to_db(fd):
    """ convert a FileDivision to a Division """
    args, _ = Division.subtypes_from_id(fd.id)
    if fd.sameAs:
        args["redirect_id"] = fd.sameAs
    return Division(id=fd.id, name=fd.name, **args)


def load_divisions(country, bulk=False):
    existing_divisions = Division.objects.filter(country=country)

    country_division = FileDivision.get("ocd-division/country:{}".format(country))
    objects = [to_db(country_division)]

    for child in country_division.children(levels=100):
        objects.append(to_db(child))

    print(
        "{} divisions found in the CSV, and {} already in the DB".format(
            len(objects), existing_divisions.count()
        )
    )

    if set(objects) == set(existing_divisions):
        print("The CSV and the DB contents are exactly the same; no work to be done!")
    else:
        if bulk:
            # delete old ids and add new ones all at once
            with transaction.atomic():
                Division.objects.filter(country=country).delete()
                Division.objects.bulk_create(objects, batch_size=10000)
            print("{} divisions created".format(len(objects)))
        else:
            to_create = set(objects) - set(existing_divisions)
            to_delete = set(existing_divisions) - set(objects)
            # delete removed ids and add new ones all at once
            with transaction.atomic():
                for division in to_delete:
                    division.delete()
                for division in to_create:
                    division.save()
            print("{} divisions deleted".format(len(to_delete)))
            print("{} divisions created".format(len(to_create)))


class Command(BaseCommand):
    help = "initialize divisions"

    def add_arguments(self, parser):
        parser.add_argument("countries", nargs="+", type=str)
        parser.add_argument(
            "--bulk",
            action="store_true",
            help="Use bulk_create to add divisions. *Warning* This deletes any existing divisions",
        )

    def handle(self, *args, **options):
        for country in options["countries"]:
            load_divisions(country, options["bulk"])



from django.contrib import admin
from django.apps import apps
from .. import models
from .base import ModelAdmin, ReadOnlyTabularInline, ContactDetailInline, LinkInline


@admin.register(models.Division)
class DivisionAdmin(ModelAdmin):
    list_display = ("name", "id")
    search_fields = list_display
    fields = readonly_fields = ("id", "name", "redirect", "country")
    ordering = ("id",)


# have to handle this special since LegislativeSession might not be present
try:
    LegislativeSession = apps.get_model("legislative", "LegislativeSession")

    class LegislativeSessionInline(ReadOnlyTabularInline):
        model = LegislativeSession
        readonly_fields = (
            "identifier",
            "name",
            "classification",
            "start_date",
            "end_date",
        )
        ordering = ("-identifier",)

    jurisdiction_inlines = [LegislativeSessionInline]

except LookupError:
    jurisdiction_inlines = []


@admin.register(models.Jurisdiction)
class JurisdictionAdmin(ModelAdmin):
    list_display = ("name", "id")
    readonly_fields = fields = (
        "id",
        "name",
        "division",
        "classification",
        "feature_flags",
        "extras",
        "url",
    )
    ordering = ("id",)
    inlines = jurisdiction_inlines


class PostContactDetailInline(ContactDetailInline):
    model = models.PostContactDetail


class PostLinkInline(LinkInline):
    model = models.PostLink


@admin.register(models.Post)
class PostAdmin(ModelAdmin):
    readonly_fields = ("id", "label", "organization", "division", "extras", "role")
    fields = readonly_fields + (("start_date", "end_date"),)
    list_display = ("label", "organization", "division")
    list_filter = ("organization__jurisdiction__name",)
    ordering = ("organization__name",)
    inlines = [PostContactDetailInline, PostLinkInline]
    search_fields = ("organization__name", "label")



import datetime
from django.db import models
from django.db.models import Q, QuerySet
from .base import OCDBase, LinkBase, OCDIDField, RelatedBase, IdentifierBase
from .division import Division
from .jurisdiction import Jurisdiction
from .. import common

# abstract models


class ContactDetailBase(RelatedBase):
    """
    A base class for ContactDetail models.
    """

    type = models.CharField(
        max_length=50,
        choices=common.CONTACT_TYPE_CHOICES,
        help_text="The type of Contact being defined.",
    )
    value = models.CharField(
        max_length=300,
        help_text="The content of the Contact information like a phone number or email address.",
    )
    note = models.CharField(
        max_length=300,
        blank=True,
        help_text="A short, optional note about the Contact value.",
    )
    label = models.CharField(
        max_length=300, blank=True, help_text="A title for the content of the Contact."
    )

    class Meta:
        abstract = True

    def __str__(self):
        return "{}: {}".format(self.get_type_display(), self.value)


class OtherNameBase(RelatedBase):
    """
    A base class for OtherName models.
    """

    name = models.CharField(
        max_length=500, db_index=True, help_text="An alternative name."
    )
    note = models.CharField(
        max_length=500,
        blank=True,
        help_text="A short, optional note about alternative name.",
    )
    start_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="An optional start date for usage of the alternative name "
        "in YYYY[-MM[-DD]] string format.",
    )
    end_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="An optional end date for usage of the alternative name in "
        "YYYY[-MM[-DD]] string format.",
    )

    class Meta:
        abstract = True

    def __str__(self):
        return "{} ({})".format(self.name, self.note)


# the actual models


class Organization(OCDBase):
    """
    A group of people, typically in a legislative or rule-making context.
    """

    id = OCDIDField(ocd_type="organization")
    name = models.CharField(max_length=300, help_text="The name of the Organization.")
    image = models.URLField(
        blank=True,
        max_length=2000,
        help_text="A URL leading to an image that identifies the Organization visually.",
    )
    parent = models.ForeignKey(
        "self",
        related_name="children",
        null=True,
        # parent can be deleted w/o affecting children
        on_delete=models.SET_NULL,
        help_text="A link to another Organization that serves as this Organization's parent.",
    )
    jurisdiction = models.ForeignKey(
        Jurisdiction,
        related_name="organizations",
        null=True,
        # deletion of a jurisdiction should be hard
        on_delete=models.PROTECT,
        help_text="A link to the Jurisdiction that contains this Organization.",
    )
    classification = models.CharField(
        max_length=100,
        blank=True,
        choices=common.ORGANIZATION_CLASSIFICATION_CHOICES,
        help_text="The type of Organization being defined.",
    )
    founding_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The founding date of the Organization in YYYY[-MM[-DD]] string format.",
    )
    dissolution_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The dissolution date of the Organization in YYYY[-MM[-DD]] string format.",
    )

    def __str__(self):
        return self.name

    # Access all "ancestor" organizations
    def get_parents(self):
        org = self
        while True:
            org = org.parent
            # Django accesses parents lazily, so have to check if one actually exists
            if org:
                yield org
            else:
                break

    def get_current_members(self):
        """ return all Person objects w/ current memberships to org """
        today = datetime.date.today().isoformat()

        return Person.objects.filter(
            Q(memberships__start_date="") | Q(memberships__start_date__lte=today),
            Q(memberships__end_date="") | Q(memberships__end_date__gte=today),
            memberships__organization_id=self.id,
        )

    class Meta:
        db_table = "opencivicdata_organization"
        index_together = [
            ["jurisdiction", "classification", "name"],
            ["classification", "name"],
        ]


class OrganizationIdentifier(IdentifierBase):
    """
    Upstream identifiers of an Organization.
    """

    organization = models.ForeignKey(
        Organization,
        related_name="identifiers",
        help_text="Reference to the Organization identified by this alternative identifier.",
        on_delete=models.CASCADE,
    )

    def __str__(self):
        tmpl = "%s identifies %s"
        return tmpl % (self.identifier, self.organization)

    class Meta:
        db_table = "opencivicdata_organizationidentifier"


class OrganizationName(OtherNameBase):
    """
    Alternate or former name for an Organization.
    """

    organization = models.ForeignKey(
        Organization,
        related_name="other_names",
        help_text="A link to the Organization with this alternative name.",
        on_delete=models.CASCADE,
    )

    class Meta:
        db_table = "opencivicdata_organizationname"


class OrganizationContactDetail(ContactDetailBase):
    """
    Contact information for an Organization.
    """

    organization = models.ForeignKey(
        Organization,
        related_name="contact_details",
        help_text="A link to the Organization connected to this contact.",
        on_delete=models.CASCADE,
    )

    class Meta:
        db_table = "opencivicdata_organizationcontactdetail"


class OrganizationLink(LinkBase):
    """
    URL for a document about an Organization.
    """

    organization = models.ForeignKey(
        Organization,
        related_name="links",
        help_text="A reference to the Organization connected to this link.",
        on_delete=models.CASCADE,
    )

    class Meta:
        db_table = "opencivicdata_organizationlink"


class OrganizationSource(LinkBase):
    """
    Source used in assembling an Organization.
    """

    organization = models.ForeignKey(
        Organization,
        related_name="sources",
        help_text="A link to the Organization connected to this source.",
        on_delete=models.CASCADE,
    )

    class Meta:
        db_table = "opencivicdata_organizationsource"


class Post(OCDBase):
    """
    A position in an organization that exists independently of the person holding it.
    """

    id = OCDIDField(ocd_type="post")
    label = models.CharField(max_length=300, help_text="A label describing the Post.")
    role = models.CharField(
        max_length=300,
        blank=True,
        help_text="The function that the holder of the post fulfills.",
    )
    organization = models.ForeignKey(
        Organization,
        related_name="posts",
        help_text="The Organization in which the post is held.",
        on_delete=models.CASCADE,
    )
    division = models.ForeignKey(
        Division,
        related_name="posts",
        null=True,
        blank=True,
        default=None,
        help_text="The Division where the post exists.",
        # if the division goes away the post is just jurisdiction-less
        on_delete=models.SET_NULL,
    )
    start_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="An optional start date for the Post in YYYY[-MM[-DD]] string format.",
    )
    end_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="An optional end date for the Post in YYYY[-MM[-DD]] string format.",
    )
    maximum_memberships = models.PositiveIntegerField(
        default=1, help_text="The maximum number of people who can hold this Post."
    )

    class Meta:
        db_table = "opencivicdata_post"
        index_together = [["organization", "label"]]

    def __str__(self):
        return "{} - {} - {}".format(self.role, self.label, self.organization)


class PostContactDetail(ContactDetailBase):
    """
    Contact information for whoever currently occupies a Post.
    """

    post = models.ForeignKey(
        Post,
        related_name="contact_details",
        help_text="A link to the Post connected to this contact.",
        on_delete=models.CASCADE,
    )

    class Meta:
        db_table = "opencivicdata_postcontactdetail"


class PostLink(LinkBase):
    """
    URL for a document about a Post.
    """

    post = models.ForeignKey(
        Post,
        related_name="links",
        on_delete=models.CASCADE,
        help_text="A reference to the Post connected to this link.",
    )

    class Meta:
        db_table = "opencivicdata_postlink"


class PersonQuerySet(QuerySet):
    def member_of(self, organization_name, current_only=True, post=None):
        filter_params = []

        if current_only:
            today = datetime.date.today().isoformat()

            filter_params = [
                Q(memberships__start_date="") | Q(memberships__start_date__lte=today),
                Q(memberships__end_date="") | Q(memberships__end_date__gte=today),
            ]
        if post:
            filter_params.append(Q(memberships__post__label=post))

        if organization_name.startswith("ocd-organization/"):
            qs = self.filter(
                *filter_params, memberships__organization_id=organization_name
            )
        else:
            qs = self.filter(
                *filter_params, memberships__organization__name=organization_name
            )
        return qs


class Person(OCDBase):
    """
    An individual that has served in a political office.
    """

    objects = PersonQuerySet.as_manager()

    id = OCDIDField(ocd_type="person")
    name = models.CharField(
        max_length=300, db_index=True, help_text="A Person's preferred full name."
    )
    sort_name = models.CharField(
        max_length=100,
        default="",
        blank=True,
        help_text="A version of a Person's full name rearranged for alphabetical sorting.",
    )
    family_name = models.CharField(
        max_length=100, blank=True, help_text="A Person's family name."
    )
    given_name = models.CharField(
        max_length=100, blank=True, help_text="A Person's given name."
    )
    image = models.URLField(
        blank=True,
        max_length=2000,
        help_text="A URL leading to an image that identifies the Person visually.",
    )
    gender = models.CharField(max_length=100, blank=True, help_text="A Person's gender")
    summary = models.CharField(
        max_length=500,
        blank=True,
        help_text="A short, one-line account of a Person's life.",
    )
    national_identity = models.CharField(
        max_length=300, blank=True, help_text="The nation a Person is identified with."
    )
    biography = models.TextField(
        blank=True, help_text="An extended account of a Person's life."
    )
    birth_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date of a Person's birth in YYYY[-MM[-DD]] string format.",
    )
    death_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date of a Person's death in YYYY[-MM[-DD]] string format.",
    )

    def __str__(self):
        return self.name

    def add_other_name(self, name, note=""):
        PersonName.objects.create(name=name, note=note, person_id=self.id)

    class Meta:
        db_table = "opencivicdata_person"
        verbose_name_plural = "people"


class PersonIdentifier(IdentifierBase):
    """
    Upstream identifier for a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="identifiers",
        on_delete=models.CASCADE,
        help_text="A link to the Person connected to this alternative identifier.",
    )

    class Meta:
        db_table = "opencivicdata_personidentifier"


class PersonName(OtherNameBase):
    """
    Alternate or former name of a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="other_names",
        on_delete=models.CASCADE,
        help_text="A link to the Person connected to this alternative name.",
    )

    class Meta:
        db_table = "opencivicdata_personname"


class PersonContactDetail(ContactDetailBase):
    """
    Contact information for a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="contact_details",
        on_delete=models.CASCADE,
        help_text="A link to the Person connected to this contact.",
    )

    class Meta:
        db_table = "opencivicdata_personcontactdetail"


class PersonLink(LinkBase):
    """
    URL for a document about a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="links",
        on_delete=models.CASCADE,
        help_text="A reference to the Person connected to this link.",
    )

    class Meta:
        db_table = "opencivicdata_personlink"


class PersonSource(LinkBase):
    """
    Source used in assembling a Person.
    """

    person = models.ForeignKey(
        Person,
        related_name="sources",
        on_delete=models.CASCADE,
        help_text="A link to the Person connected to this source.",
    )

    class Meta:
        db_table = "opencivicdata_personsource"


class Membership(OCDBase):
    """
    A relationship between a Person and an Organization, possibly including a Post.
    """

    id = OCDIDField(ocd_type="membership")
    organization = models.ForeignKey(
        Organization,
        related_name="memberships",
        # memberships will go away if the org does
        on_delete=models.CASCADE,
        help_text="A link to the Organization in which the Person is a member.",
    )
    person = models.ForeignKey(
        Person,
        related_name="memberships",
        null=True,
        # Membership will just unlink if the person goes away
        on_delete=models.SET_NULL,
        help_text="A link to the Person that is a member of the Organization.",
    )
    person_name = models.CharField(
        max_length=300,
        blank=True,
        default="",
        help_text="The name of the Person that is a member of the Organization.",
    )
    post = models.ForeignKey(
        Post,
        related_name="memberships",
        null=True,
        # Membership will just unlink if the post goes away
        on_delete=models.SET_NULL,
        help_text="	The Post held by the member in the Organization.",
    )
    on_behalf_of = models.ForeignKey(
        Organization,
        related_name="memberships_on_behalf_of",
        null=True,
        help_text="The Organization on whose behalf the Person is a member of the Organization.",
        # Membership will just unlink if the org goes away
        on_delete=models.SET_NULL,
    )
    label = models.CharField(
        max_length=300, blank=True, help_text="A label describing the membership."
    )
    role = models.CharField(
        max_length=300,
        blank=True,
        help_text="The role that the member fulfills in the Organization.",
    )
    start_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date on which the relationship began in YYYY[-MM[-DD]] string format.",
    )
    end_date = models.CharField(
        max_length=10,
        blank=True,
        help_text="The date on which the relationship ended in YYYY[-MM[-DD]] string format.",
    )

    class Meta:
        db_table = "opencivicdata_membership"
        index_together = [["organization", "person", "label", "post"]]

    def __str__(self):
        return "{} in {} ({})".format(self.person, self.organization, self.role)


class MembershipContactDetail(ContactDetailBase):
    """
    Contact information for Person at an Organization.
    """

    membership = models.ForeignKey(
        Membership,
        related_name="contact_details",
        help_text="A link to the Membership connected to this contact.",
        on_delete=models.CASCADE,
    )

    class Meta:
        db_table = "opencivicdata_membershipcontactdetail"


class MembershipLink(LinkBase):
    """
    URL for a document about a Person's relationship to an Organization.
    """

    membership = models.ForeignKey(
        Membership,
        related_name="links",
        help_text="A reference to the Membership connected to this link.",
        on_delete=models.CASCADE,
    )

    class Meta:
        db_table = "opencivicdata_membershiplink"



import pytest
from openstates.scrape import Membership as ScrapeMembership
from openstates.scrape import Person as ScrapePerson
from openstates.importers import (
    MembershipImporter,
    PersonImporter,
    OrganizationImporter,
)
from openstates.exceptions import NoMembershipsError
from openstates.data.models import (
    Organization,
    Post,
    Person,
    Division,
    Jurisdiction,
)


class DumbMockImporter(object):
    """ this is a mock importer that implements a resolve_json_id that is just a pass-through """

    json_to_db_id = {}

    def resolve_json_id(self, json_id, allow_no_match=False):
        return json_id


def create_jurisdiction():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    Jurisdiction.objects.create(id="fnd-jid", division_id="ocd-division/country:us")


@pytest.mark.django_db
def test_full_membership():
    create_jurisdiction()
    org = Organization.objects.create(
        id="fnd",
        name="Foundation",
        classification="foundation",
        jurisdiction_id="fnd-jid",
    )
    hari = Person.objects.create(id="hs", name="Hari Seldon")
    robot = Person.objects.create(id="robot", name="R. Daneel Olivaw")
    post = Post.objects.create(
        id="f", label="founder", role="Founder", organization=org
    )

    # add a membership through a post
    m1 = ScrapeMembership(person_id=hari.id, organization_id=org.id, post_id=post.id)
    m1.add_contact_detail(type="phone", value="555-555-1234", note="this is fake")
    m1.add_link("http://example.com/link")

    # add a membership direct to an organization
    m2 = ScrapeMembership(
        person_id=robot.id, organization_id=org.id, label="member", role="member"
    )

    dumb_imp = DumbMockImporter()
    memimp = MembershipImporter("fnd-jid", dumb_imp, dumb_imp, dumb_imp)
    memimp.import_data([m1.as_dict(), m2.as_dict()])

    # ensure that the memberships attached in the right places
    assert org.memberships.count() == 2
    assert hari.memberships.count() == 1
    assert robot.memberships.count() == 1
    assert post.memberships.count() == 1

    # ensure that the first membership has contact details and links
    m = hari.memberships.get()
    cd = m.contact_details.get()
    assert cd.type == "phone"
    assert cd.value == "555-555-1234"
    assert cd.note == "this is fake"
    assert m.links.all()[0].url == "http://example.com/link"


@pytest.mark.django_db
def test_no_membership_for_person():
    create_jurisdiction()
    Organization.objects.create(
        id="fnd",
        name="Foundation",
        classification="foundation",
        jurisdiction_id="fnd-jid",
    )

    # import a person with no memberships
    p = ScrapePerson("a man without a country")
    person_imp = PersonImporter("fnd-jid")
    person_imp.import_data([p.as_dict()])

    # try to import a membership
    dumb_imp = DumbMockImporter()
    memimp = MembershipImporter("fnd-jid", person_imp, dumb_imp, dumb_imp)

    with pytest.raises(NoMembershipsError):
        memimp.import_data([])


@pytest.mark.django_db
def test_no_membership_for_person_including_party():
    """
    even though party is specified we should still get a no memberships error because it doesn't
    bind the person to a jurisdiction, thus causing duplication
    """
    create_jurisdiction()
    Organization.objects.create(
        id="fnd",
        name="Foundation",
        classification="foundation",
        jurisdiction_id="fnd-jid",
    )
    Organization.objects.create(id="dem", name="Democratic", classification="party")

    # import a person with no memberships
    p = ScrapePerson("a man without a country", party="Democratic")
    person_imp = PersonImporter("fnd-jid")
    org_imp = OrganizationImporter("fnd-jid")
    person_imp.import_data([p.as_dict()])

    # try to import a membership
    dumb_imp = DumbMockImporter()
    memimp = MembershipImporter("fnd-jid", person_imp, org_imp, dumb_imp)

    with pytest.raises(NoMembershipsError):
        memimp.import_data([p._related[0].as_dict()])


@pytest.mark.django_db
def test_multiple_orgs_of_same_class():
    """
    We should be able to set memberships on organizations with the
    same classification within the same jurisdictions
    """
    create_jurisdiction()
    Organization.objects.create(
        id="fnd",
        name="Foundation",
        classification="foundation",
        jurisdiction_id="fnd-jid",
    )
    Organization.objects.create(
        id="fdr",
        name="Federation",
        classification="foundation",
        jurisdiction_id="fnd-jid",
    )

    hari = ScrapePerson(
        "Hari Seldon",
        primary_org="foundation",
        role="founder",
        primary_org_name="Foundation",
    )

    picard = ScrapePerson(
        "Jean Luc Picard",
        primary_org="foundation",
        role="founder",
        primary_org_name="Federation",
    )

    person_imp = PersonImporter("fnd-jid")
    person_imp.import_data([hari.as_dict()])
    person_imp.import_data([picard.as_dict()])

    # try to import a membership
    org_imp = OrganizationImporter("fnd-jid")
    dumb_imp = DumbMockImporter()
    memimp = MembershipImporter("fnd-jid", person_imp, org_imp, dumb_imp)

    memimp.import_data([hari._related[0].as_dict(), picard._related[0].as_dict()])

    assert (
        Person.objects.get(name="Hari Seldon").memberships.get().organization.name
        == "Foundation"
    )
    assert (
        Person.objects.get(name="Jean Luc Picard").memberships.get().organization.name
        == "Federation"
    )


@pytest.mark.django_db
def test_multiple_posts_class():
    create_jurisdiction()

    org = Organization.objects.create(
        id="fnd",
        name="Foundation",
        classification="foundation",
        jurisdiction_id="fnd-jid",
    )
    hari = Person.objects.create(id="hs", name="Hari Seldon")
    founder = Post.objects.create(
        id="f", label="founder", role="Founder", organization=org
    )
    chair = Post.objects.create(id="c", label="chair", role="Chair", organization=org)

    m1 = ScrapeMembership(person_id=hari.id, organization_id=org.id, post_id=founder.id)
    m2 = ScrapeMembership(person_id=hari.id, organization_id=org.id, post_id=chair.id)

    dumb_imp = DumbMockImporter()
    memimp = MembershipImporter("fnd-jid", dumb_imp, dumb_imp, dumb_imp)
    memimp.import_data([m1.as_dict(), m2.as_dict()])

    # ensure that the memberships attached in the right places
    assert org.memberships.count() == 2
    assert hari.memberships.count() == 2
    assert founder.memberships.count() == 1
    assert chair.memberships.count() == 1


@pytest.mark.django_db
def test_unmatched_person():
    create_jurisdiction()

    org = Organization.objects.create(
        id="fnd",
        name="Foundation",
        classification="foundation",
        jurisdiction_id="fnd-jid",
    )
    # not a real person, won't have a person_id after import
    m1 = ScrapeMembership(
        person_name="Harry Seldom", organization_id=org.id, person_id=None
    )

    dumb_imp = DumbMockImporter()
    memimp = MembershipImporter("fnd-jid", dumb_imp, dumb_imp, dumb_imp)
    memimp.import_data([m1.as_dict()])

    # ensure that the memberships attached in the right places
    assert org.memberships.count() == 1

    membership = org.memberships.get()
    assert membership.person_id is None
    assert membership.person_name == "Harry Seldom"



import os
import copy
import glob
import json
import logging
from django.db.models import Q
from django.db.models.signals import post_save
from django.contrib.contenttypes.models import ContentType
from .. import settings
from ..data.models import LegislativeSession
from ..exceptions import DuplicateItemError, UnresolvedIdError, DataImportError
from ..reports.models import Identifier
from ..utils import get_pseudo_id, utcnow


def omnihash(obj):
    """ recursively hash unhashable objects """
    if isinstance(obj, set):
        return hash(frozenset(omnihash(e) for e in obj))
    elif isinstance(obj, (tuple, list)):
        return hash(tuple(omnihash(e) for e in obj))
    elif isinstance(obj, dict):
        return hash(frozenset((k, omnihash(v)) for k, v in obj.items()))
    else:
        return hash(obj)


def items_differ(jsonitems, dbitems, subfield_dict):
    """ check whether or not jsonitems and dbitems differ """

    # short circuit common cases
    if len(jsonitems) == len(dbitems) == 0:
        # both are empty
        return False
    elif len(jsonitems) != len(dbitems):
        # if lengths differ, they're definitely different
        return True

    original_jsonitems = jsonitems
    jsonitems = copy.deepcopy(jsonitems)
    keys = jsonitems[0].keys()

    # go over dbitems looking for matches
    for dbitem in dbitems:
        order = getattr(dbitem, "order", None)
        match = None
        for i, jsonitem in enumerate(jsonitems):
            # check if all keys (excluding subfields) match
            for k in keys:
                if k not in subfield_dict and getattr(dbitem, k) != jsonitem.get(
                    k, None
                ):
                    break
            else:
                # all fields match so far, possibly equal, just check subfields now
                for k in subfield_dict:
                    jsonsubitems = jsonitem[k]
                    dbsubitems = list(getattr(dbitem, k).all())
                    if items_differ(jsonsubitems, dbsubitems, subfield_dict[k][2]):
                        break
                else:
                    # if the dbitem sets 'order', then the order matters
                    if order is not None and int(order) != original_jsonitems.index(
                        jsonitem
                    ):
                        break
                    # these items are equal, so let's mark it for removal
                    match = i
                    break

        if match is not None:
            # item exists in both, remove from jsonitems
            jsonitems.pop(match)
        else:
            # exists in db but not json
            return True

    # if we get here, jsonitems has to be empty because we asserted that the length was
    # the same and we found a match for each thing in dbitems, here's a safety check just in case
    if jsonitems:  # pragma: no cover
        return True

    return False


class BaseImporter(object):
    """ BaseImporter

    Override:
        get_object(data)
        limit_spec(spec)                [optional, required if pseudo_ids are used]
        prepare_for_db(data)            [optional]
        postimport()                    [optional]
    """

    _type = None
    model_class = None
    related_models = {}
    preserve_order = set()
    merge_related = {}
    cached_transformers = {}

    def __init__(self, jurisdiction_id):
        self.jurisdiction_id = jurisdiction_id
        self.json_to_db_id = {}
        self.duplicates = {}
        self.pseudo_id_cache = {}
        self.session_cache = {}
        self.logger = logging.getLogger("openstates")
        self.info = self.logger.info
        self.debug = self.logger.debug
        self.warning = self.logger.warning
        self.error = self.logger.error
        self.critical = self.logger.critical

        # load transformers from appropriate setting
        if settings.IMPORT_TRANSFORMERS.get(self._type):
            self.cached_transformers = settings.IMPORT_TRANSFORMERS[self._type]

    def get_session_id(self, identifier):
        if identifier not in self.session_cache:
            self.session_cache[identifier] = LegislativeSession.objects.get(
                identifier=identifier, jurisdiction_id=self.jurisdiction_id
            ).id
        return self.session_cache[identifier]

    # no-ops to be overriden
    def prepare_for_db(self, data):
        return data

    def postimport(self):
        pass

    def resolve_json_id(self, json_id, allow_no_match=False):
        """
            Given an id found in scraped JSON, return a DB id for the object.

            params:
                json_id:        id from json
                allow_no_match: just return None if id can't be resolved

            returns:
                database id

            raises:
                ValueError if id couldn't be resolved
        """
        if not json_id:
            return None

        if json_id.startswith("~"):
            # keep caches of all the pseudo-ids to avoid doing 1000s of lookups during import
            if json_id not in self.pseudo_id_cache:
                spec = get_pseudo_id(json_id)
                spec = self.limit_spec(spec)

                if isinstance(spec, Q):
                    objects = self.model_class.objects.filter(spec)
                else:
                    objects = self.model_class.objects.filter(**spec)
                ids = {each.id for each in objects}
                if len(ids) == 1:
                    self.pseudo_id_cache[json_id] = ids.pop()
                    errmsg = None
                elif not ids:
                    errmsg = "cannot resolve pseudo id to {}: {}".format(
                        self.model_class.__name__, json_id
                    )
                else:
                    errmsg = "multiple objects returned for {} pseudo id {}: {}".format(
                        self.model_class.__name__, json_id, ids
                    )

                # either raise or log error
                if errmsg:
                    if not allow_no_match:
                        raise UnresolvedIdError(errmsg)
                    else:
                        self.error(errmsg)
                        self.pseudo_id_cache[json_id] = None

            # return the cached object
            return self.pseudo_id_cache[json_id]

        # get the id that the duplicate points to, or use self
        json_id = self.duplicates.get(json_id, json_id)

        try:
            return self.json_to_db_id[json_id]
        except KeyError:
            raise UnresolvedIdError("cannot resolve id: {}".format(json_id))

    def import_directory(self, datadir):
        """ import a JSON directory into the database """

        def json_stream():
            # load all json, mapped by json_id
            for fname in glob.glob(os.path.join(datadir, self._type + "_*.json")):
                with open(fname) as f:
                    yield json.load(f)

        return self.import_data(json_stream())

    def _prepare_imports(self, dicts):

        """ filters the import stream to remove duplicates

        also serves as a good place to override if anything special has to be done to the
        order of the import stream (see OrganizationImporter)
        """
        # hash(json): id
        seen_hashes = {}

        for data in dicts:
            json_id = data.pop("_id")

            # map duplicates (using omnihash to tell if json dicts are identical-ish)
            objhash = omnihash(data)
            if objhash not in seen_hashes:
                seen_hashes[objhash] = json_id
                yield json_id, data
            else:
                self.duplicates[json_id] = seen_hashes[objhash]

    def import_data(self, data_items):
        """ import a bunch of dicts together """
        # keep counts of all actions
        record = {
            "insert": 0,
            "update": 0,
            "noop": 0,
            "start": utcnow(),
            "records": {"insert": [], "update": [], "noop": []},
        }

        for json_id, data in self._prepare_imports(data_items):
            obj_id, what = self.import_item(data)
            self.json_to_db_id[json_id] = obj_id
            record["records"][what].append(obj_id)
            record[what] += 1

        # all objects are loaded, a perfect time to do inter-object resolution and other tasks
        self.postimport()

        record["end"] = utcnow()

        return {self._type: record}

    def import_item(self, data):
        """ function used by import_data """
        what = "noop"

        # remove the JSON _id (may still be there if called directly)
        data.pop("_id", None)

        # add fields/etc.
        data = self.apply_transformers(data)
        data = self.prepare_for_db(data)

        try:
            obj = self.get_object(data)
        except self.model_class.DoesNotExist:
            obj = None

        # remove pupa_id which does not belong in the OCD data models
        pupa_id = data.pop("pupa_id", None)

        # pull related fields off
        related = {}
        for field in self.related_models:
            related[field] = data.pop(field)

        # obj existed, check if we need to do an update
        if obj:
            if obj.id in self.json_to_db_id.values():
                raise DuplicateItemError(data, obj, related.get("sources", []))
            # check base object for changes
            for key, value in data.items():
                if getattr(obj, key) != value and key not in obj.locked_fields:
                    setattr(obj, key, value)
                    what = "update"

            updated = self._update_related(obj, related, self.related_models)
            if updated:
                what = "update"

            if what == "update":
                obj.save()

        # need to create the data
        else:
            what = "insert"
            try:
                obj = self.model_class.objects.create(**data)
            except Exception as e:
                raise DataImportError(
                    "{} while importing {} as {}".format(e, data, self.model_class)
                )
            self._create_related(obj, related, self.related_models)

            # Fire post-save signal after related objects are created to allow
            # for handlers make use of related objects
            post_save.send(sender=self.model_class, instance=obj, created=True)

        if pupa_id:
            Identifier.objects.get_or_create(
                identifier=pupa_id,
                jurisdiction_id=self.jurisdiction_id,
                defaults={"content_object": obj},
            )

        return obj.id, what

    def _update_related(self, obj, related, subfield_dict):
        """
        update DB objects related to a base object
            obj:            a base object to create related
            related:        dict mapping field names to lists of related objects
            subfield_list:  where to get the next layer of subfields
        """
        # keep track of whether or not anything was updated
        updated = False

        # for each related field - check if there are differences
        for field, items in related.items():
            # skip subitem check if it's locked anyway
            if field in obj.locked_fields:
                continue

            # get items from database
            dbitems = list(getattr(obj, field).all())
            dbitems_count = len(dbitems)

            # default to doing nothing
            do_delete = do_update = False

            if items and dbitems_count:  # we have items, so does db, check for conflict
                do_delete = do_update = items_differ(
                    items, dbitems, subfield_dict[field][2]
                )
            elif items and not dbitems_count:  # we have items, db doesn't, just update
                do_update = True
            elif not items and dbitems_count:  # db has items, we don't, just delete
                do_delete = True
            # otherwise: no items or dbitems, so nothing is done

            # don't delete if field is in merge_related
            if field in self.merge_related:
                new_items = []
                # build a list of keyfields to existing database objects
                keylist = self.merge_related[field]
                keyed_dbitems = {
                    tuple(getattr(item, k) for k in keylist): item for item in dbitems
                }

                # go through 'new' items
                #   if item with the same keyfields exists:
                #       update the database item w/ the new item's properties
                #   else:
                #       add it to new_items
                for item in items:
                    key = tuple(item.get(k) for k in keylist)
                    dbitem = keyed_dbitems.get(key)
                    if not dbitem:
                        new_items.append(item)
                    else:
                        # update dbitem
                        for fname, val in item.items():
                            setattr(dbitem, fname, val)
                        dbitem.save()

                # import anything that made it to new_items in the usual fashion
                self._create_related(obj, {field: new_items}, subfield_dict)
            else:
                # default logic is to just wipe and recreate subobjects
                if do_delete:
                    updated = True
                    getattr(obj, field).all().delete()
                if do_update:
                    updated = True
                    self._create_related(obj, {field: items}, subfield_dict)

        return updated

    def _create_related(self, obj, related, subfield_dict):
        """
        create DB objects related to a base object
            obj:            a base object to create related
            related:        dict mapping field names to lists of related objects
            subfield_list:  where to get the next layer of subfields
        """
        for field, items in related.items():
            subobjects = []
            all_subrelated = []
            Subtype, reverse_id_field, subsubdict = subfield_dict[field]
            for order, item in enumerate(items):
                # pull off 'subrelated' (things that are related to this obj)
                subrelated = {}
                for subfield in subsubdict:
                    subrelated[subfield] = item.pop(subfield)

                if field in self.preserve_order:
                    item["order"] = order

                item[reverse_id_field] = obj.id

                try:
                    subobjects.append(Subtype(**item))
                    all_subrelated.append(subrelated)
                except Exception as e:
                    raise DataImportError(
                        "{} while importing {} as {}".format(e, item, Subtype)
                    )

            # add all subobjects at once (really great for actions & votes)
            try:
                Subtype.objects.bulk_create(subobjects)
            except Exception as e:
                raise DataImportError(
                    "{} while importing {} as {}".format(e, subobjects, Subtype)
                )

            # after import the subobjects, import their subsubobjects
            for subobj, subrel in zip(subobjects, all_subrelated):
                self._create_related(subobj, subrel, subsubdict)

    def lookup_obj_id(self, pupa_id, model):
        content_type = ContentType.objects.get_for_model(model)
        try:
            obj_id = Identifier.objects.get(
                identifier=pupa_id,
                content_type=content_type,
                jurisdiction_id=self.jurisdiction_id,
            ).object_id
        except Identifier.DoesNotExist:
            obj_id = None

        return obj_id

    def apply_transformers(self, data, transformers=None):
        if transformers is None:
            transformers = self.cached_transformers

        for key, key_transformers in transformers.items():
            if key not in data:
                continue
            if isinstance(key_transformers, list):
                for transformer in key_transformers:
                    data[key] = transformer(data[key])
            elif isinstance(key_transformers, dict):
                self.apply_transformers(data[key], key_transformers)
            else:
                data[key] = key_transformers(data[key])

        return data

    def get_seen_sessions(self):
        return self.session_cache.values()



from django.urls import reverse
from django.contrib import admin
from .. import models
from .base import (
    ModelAdmin,
    ReadOnlyTabularInline,
    IdentifierInline,
    LinkInline,
    ContactDetailInline,
    OtherNameInline,
)


class OrganizationIdentifierInline(IdentifierInline):
    model = models.OrganizationIdentifier


class OrganizationNameInline(OtherNameInline):
    model = models.OrganizationName


class OrganizationContactDetailInline(ContactDetailInline):
    model = models.OrganizationContactDetail


class OrganizationLinkInline(LinkInline):
    model = models.OrganizationLink


class OrganizationSourceInline(LinkInline):
    model = models.OrganizationSource


class PostInline(admin.TabularInline):
    """ a read-only inline for posts here, with links to the real thing """

    model = models.Post
    extra = 0
    fields = readonly_fields = ("label", "role")
    ordering = ("label",)
    can_delete = False
    show_change_link = True

    def has_add_permission(self, request):
        return False


class OrgMembershipInline(ReadOnlyTabularInline):
    model = models.Membership
    fk_name = "organization"
    readonly_fields = ("id", "person", "post", "label", "role", "start_date")
    fields = readonly_fields + ("end_date",)
    extra = 0
    can_delete = False


@admin.register(models.Organization)
class OrganizationAdmin(ModelAdmin):
    readonly_fields = (
        "id",
        "name",
        "classification",
        "parent",
        "jurisdiction",
        "extras",
    )
    fields = readonly_fields + (("founding_date", "dissolution_date"), "image")
    search_fields = ("name",)
    list_filter = ("jurisdiction__name",)

    inlines = [
        OrganizationIdentifierInline,
        OrganizationNameInline,
        OrganizationContactDetailInline,
        OrganizationLinkInline,
        OrganizationSourceInline,
        PostInline,
        OrgMembershipInline,
    ]

    def get_org_name(self, obj):
        parent = obj.parent
        if parent:
            return "{org} ({parent})".format(org=obj.name, parent=parent.name)
        return obj.name

    get_org_name.short_description = "Name"
    get_org_name.allow_tags = True
    get_org_name.admin_order_field = "name"

    def get_jurisdiction(self, obj):
        jurisdiction = obj.jurisdiction
        if jurisdiction:
            admin_url = reverse(
                "admin:core_jurisdiction_change", args=(jurisdiction.pk,)
            )
            tmpl = '<a href="%s">%s</a>'
            return tmpl % (admin_url, jurisdiction.name)

        return "(none)"

    get_jurisdiction.short_description = "Jurisdiction"
    get_jurisdiction.allow_tags = True
    get_jurisdiction.admin_order_field = "jurisdiction__name"

    list_select_related = ("jurisdiction",)
    list_display = ("get_org_name", "get_jurisdiction", "classification")
    ordering = ("name",)



from django.db import models
from .base import RelatedBase
from .jurisdiction import Jurisdiction
from ..common import SESSION_CLASSIFICATION_CHOICES


class LegislativeSession(RelatedBase):
    jurisdiction = models.ForeignKey(
        Jurisdiction,
        related_name="legislative_sessions",
        # should be hard to delete Jurisdiction
        on_delete=models.PROTECT,
    )
    identifier = models.CharField(max_length=100)
    name = models.CharField(max_length=300)
    classification = models.CharField(
        max_length=100, choices=SESSION_CLASSIFICATION_CHOICES, blank=True
    )
    start_date = models.CharField(max_length=10)  # YYYY[-MM[-DD]]
    end_date = models.CharField(max_length=10)  # YYYY[-MM[-DD]]

    def __str__(self):
        return "{} {}".format(self.jurisdiction, self.name)

    class Meta:
        db_table = "opencivicdata_legislativesession"



from .common import extras, fuzzy_date_blank

schema = {
    "type": "object",
    "properties": {
        "name": {"type": "string", "minLength": 1},
        "url": {"type": "string", "minLength": 1},
        "classification": {"type": "string", "minLength": 1},  # TODO: enum
        "division_id": {"type": "string", "minLength": 1},
        "legislative_sessions": {
            "type": "array",
            "items": {
                "type": "object",
                "properties": {
                    "name": {"type": "string", "minLength": 1},
                    "type": {"type": "string", "enum": ["primary", "special"]},
                    "start_date": fuzzy_date_blank,
                    "end_date": fuzzy_date_blank,
                },
            },
        },
        "feature_flags": {"type": "array", "items": {"type": "string", "minLength": 1}},
        "extras": extras,
    },
}



import pytest
from openstates.scrape import Person as ScrapePerson
from openstates.importers import PersonImporter
from openstates.data.models import (
    Person,
    Organization,
    Membership,
    Division,
    Jurisdiction,
)
from openstates.exceptions import UnresolvedIdError, SameNameError


def create_jurisdiction():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")


@pytest.mark.django_db
def test_full_person():
    person = ScrapePerson("Tom Sawyer")
    person.add_identifier("1")
    person.add_name("Tommy", start_date="1880")
    person.add_contact_detail(type="phone", value="555-555-1234", note="this is fake")
    person.add_link("http://example.com/link")
    person.add_source("http://example.com/source")

    # import person
    pd = person.as_dict()
    PersonImporter("jid").import_data([pd])

    # get person from db and assert it imported correctly
    p = Person.objects.get()
    assert "ocd-person" in p.id
    assert p.name == person.name

    assert p.identifiers.all()[0].identifier == "1"
    assert p.identifiers.all()[0].scheme == ""

    assert p.other_names.all()[0].name == "Tommy"
    assert p.other_names.all()[0].start_date == "1880"

    assert p.contact_details.all()[0].type == "phone"
    assert p.contact_details.all()[0].value == "555-555-1234"
    assert p.contact_details.all()[0].note == "this is fake"

    assert p.links.all()[0].url == "http://example.com/link"
    assert p.sources.all()[0].url == "http://example.com/source"


def create_person():
    # deduplication for people is fairly complicated, it requires a person to have a membership
    # in the jurisdiction's organization and have a matching name.  let's set that up first for
    # the deduplication tests
    p = Person.objects.create(name="Dwayne Johnson")
    p.other_names.create(name="Rocky")
    o = Organization.objects.create(name="WWE", jurisdiction_id="jid")
    Membership.objects.create(person=p, organization=o)


@pytest.mark.django_db
def test_deduplication_same_name():
    create_jurisdiction()
    create_person()
    # simplest case- just the same name
    person = ScrapePerson("Dwayne Johnson")
    pd = person.as_dict()
    PersonImporter("jid").import_data([pd])
    assert Person.objects.all().count() == 1


@pytest.mark.django_db
def test_deduplication_other_name_exists():
    create_jurisdiction()
    create_person()
    # Rocky is already saved in other_names
    person = ScrapePerson("Rocky")
    pd = person.as_dict()
    PersonImporter("jid").import_data([pd])
    assert Person.objects.all().count() == 1


@pytest.mark.django_db
def test_deduplication_other_name_overlaps():
    create_jurisdiction()
    create_person()
    # Person has other_name that overlaps w/ existing name
    person = ScrapePerson("The Rock")
    person.add_name("Dwayne Johnson")
    pd = person.as_dict()
    PersonImporter("jid").import_data([pd])
    assert Person.objects.all().count() == 1


@pytest.mark.django_db
def test_deduplication_no_name_overlap():
    create_jurisdiction()
    create_person()
    # make sure we're not just being ridiculous and avoiding importing anything in the same org
    person = ScrapePerson("CM Punk")
    pd = person.as_dict()
    PersonImporter("jid").import_data([pd])
    assert Person.objects.all().count() == 2


@pytest.mark.django_db
def test_deduplication_no_jurisdiction_overlap():
    create_jurisdiction()
    create_person()
    # make sure we get a new person if we're in a different org
    person = ScrapePerson("Dwayne Johnson")
    pd = person.as_dict()
    PersonImporter("new-jurisdiction-id").import_data([pd])
    assert Person.objects.all().count() == 2


@pytest.mark.django_db
def test_multiple_memberships():
    create_jurisdiction()
    # there was a bug where two or more memberships to the same jurisdiction
    # would cause an ORM error, this test ensures that it is fixed
    p = Person.objects.create(name="Dwayne Johnson")
    o = Organization.objects.create(name="WWE", jurisdiction_id="jid")
    Membership.objects.create(person=p, organization=o)
    o = Organization.objects.create(name="WWF", jurisdiction_id="jid")
    Membership.objects.create(person=p, organization=o)

    person = ScrapePerson("Dwayne Johnson")
    pd = person.as_dict()
    PersonImporter("jid").import_data([pd])

    # deduplication should still work
    assert Person.objects.all().count() == 1


@pytest.mark.django_db
def test_same_name_people():
    create_jurisdiction()
    o = Organization.objects.create(name="WWE", jurisdiction_id="jid")

    # importing two people with the same name to a pristine database should error
    p1 = ScrapePerson("Dwayne Johnson", image="http://example.com/1")
    p2 = ScrapePerson("Dwayne Johnson", image="http://example.com/2")
    with pytest.raises(SameNameError):
        PersonImporter("jid").import_data([p1.as_dict(), p2.as_dict()])

    # importing one person should pass
    PersonImporter("jid").import_data([p1.as_dict()])
    # create fake memberships so that future lookups work on the imported people
    for p in Person.objects.all():
        Membership.objects.create(person=p, organization=o)

    # importing another person with the same name should fail
    with pytest.raises(SameNameError):
        PersonImporter("jid").import_data([p1.as_dict(), p2.as_dict()])

    # adding birth dates should pass
    p1.birth_date = "1970"
    p2.birth_date = "1930"
    resp = PersonImporter("jid").import_data([p1.as_dict(), p2.as_dict()])
    assert resp["person"]["insert"] == 1
    assert resp["person"]["noop"] == 0
    assert resp["person"]["update"] == 1
    assert Person.objects.count() == 2
    # create fake memberships so that future lookups work on the imported people
    for p in Person.objects.all():
        Membership.objects.create(person=p, organization=o)

    # adding a third person with the same name but without a birthday should error
    p3 = ScrapePerson("Dwayne Johnson", image="http://example.com/3")

    with pytest.raises(SameNameError):
        PersonImporter("jid").import_data([p3.as_dict()])

    # and now test that an update works and we can insert a new one with the same name
    p1.image = "http://example.com/1.jpg"
    p2.birth_date = "1931"  # change birth_date, means a new insert
    resp = PersonImporter("jid").import_data([p1.as_dict(), p2.as_dict()])
    assert Person.objects.count() == 3
    assert resp["person"]["insert"] == 1
    assert resp["person"]["noop"] == 0
    assert resp["person"]["update"] == 1


@pytest.mark.django_db
def test_same_name_people_other_name():
    create_jurisdiction()
    # ensure we're taking other_names into account for the name collision code
    Organization.objects.create(name="WWE", jurisdiction_id="jid")
    p1 = ScrapePerson("Dwayne Johnson", image="http://example.com/1")
    p2 = ScrapePerson("Rock", image="http://example.com/2")
    p2.add_name("Dwayne Johnson")

    # the people have the same name but are apparently different
    with pytest.raises(SameNameError):
        PersonImporter("jid").import_data([p1.as_dict(), p2.as_dict()])


@pytest.mark.django_db
def test_same_name_second_import():
    create_jurisdiction()
    # ensure two people with the same name don't import without birthdays
    o = Organization.objects.create(name="WWE", jurisdiction_id="jid")
    p1 = ScrapePerson("Dwayne Johnson", image="http://example.com/1")
    p2 = ScrapePerson("Dwayne Johnson", image="http://example.com/2")
    p1.birth_date = "1970"
    p2.birth_date = "1930"

    # when we give them birth dates all is well though
    PersonImporter("jid").import_data([p1.as_dict(), p2.as_dict()])

    # fake some memberships so future lookups work on these people
    for p in Person.objects.all():
        Membership.objects.create(person=p, organization=o)

    p3 = ScrapePerson("Dwayne Johnson", image="http://example.com/3")

    with pytest.raises(SameNameError):
        PersonImporter("jid").import_data([p3.as_dict()])


@pytest.mark.django_db
def test_resolve_json_id():
    create_jurisdiction()
    o = Organization.objects.create(name="WWE", jurisdiction_id="jid")
    p = Person.objects.create(name="Dwayne Johnson", family_name="Johnson")
    p.other_names.create(name="Rock")
    p.memberships.create(organization=o)

    pi = PersonImporter("jid")
    assert pi.resolve_json_id('~{"name": "Dwayne Johnson"}') == p.id
    assert pi.resolve_json_id('~{"name": "Rock"}') == p.id
    assert pi.resolve_json_id('~{"name": "Johnson"}') == p.id


@pytest.mark.django_db
def test_resolve_json_id_multiple_family_name():
    create_jurisdiction()
    o = Organization.objects.create(name="WWE", jurisdiction_id="jid")
    p1 = Person.objects.create(name="Dwayne Johnson", family_name="Johnson")
    p1.other_names.create(name="Rock")
    p2 = Person.objects.create(name="Adam Johnson", family_name="Johnson")
    for p in Person.objects.all():
        Membership.objects.create(person=p, organization=o)

    # If there are multiple people with a family name, full name/other name
    # lookups should work but family name lookups should fail.
    pi = PersonImporter("jid")
    assert pi.resolve_json_id('~{"name": "Dwayne Johnson"}') == p1.id
    assert pi.resolve_json_id('~{"name": "Adam Johnson"}') == p2.id
    with pytest.raises(UnresolvedIdError):
        pi.resolve_json_id('~{"name": "Johnson"}')



import re
from ..exceptions import ScrapeValueError
from ..utils import _make_pseudo_id
from .base import BaseModel, cleanup_list, SourceMixin
from .bill import Bill
from .popolo import pseudo_organization
from .schemas.vote_event import schema


class VoteEvent(BaseModel, SourceMixin):
    _type = "vote_event"
    _schema = schema

    def __init__(
        self,
        *,
        motion_text,
        start_date,
        classification,
        result,
        legislative_session=None,
        identifier="",
        bill=None,
        bill_chamber=None,
        bill_action=None,
        organization=None,
        chamber=None
    ):
        super(VoteEvent, self).__init__()

        self.legislative_session = legislative_session
        self.motion_text = motion_text
        self.motion_classification = cleanup_list(classification, [])
        self.start_date = start_date
        self.result = result
        self.identifier = identifier
        self.bill_action = bill_action

        self.set_bill(bill, chamber=bill_chamber)

        if isinstance(bill, Bill) and not self.legislative_session:
            self.legislative_session = bill.legislative_session

        if not self.legislative_session:
            raise ScrapeValueError("must set legislative_session or bill")

        self.organization = pseudo_organization(organization, chamber, "legislature")
        self.votes = []
        self.counts = []

    def __str__(self):
        return "{0} - {1} - {2}".format(
            self.legislative_session, self.start_date, self.motion_text
        )

    def set_bill(self, bill_or_identifier, *, chamber=None):
        if not bill_or_identifier:
            self.bill = None
        elif isinstance(bill_or_identifier, Bill):
            if chamber:
                raise ScrapeValueError(
                    "set_bill takes no arguments when using a `Bill` object"
                )
            self.bill = bill_or_identifier._id
        else:
            if chamber is None:
                chamber = "legislature"
            kwargs = {
                "identifier": bill_or_identifier,
                "from_organization__classification": chamber,
                "legislative_session__identifier": self.legislative_session,
            }
            self.bill = _make_pseudo_id(**kwargs)

    def vote(self, option, voter, *, note=""):
        self.votes.append(
            {
                "option": option,
                "voter_name": voter,
                "voter_id": _make_pseudo_id(name=voter),
                "note": note,
            }
        )

    def yes(self, name, *, id=None, note=""):
        return self.vote("yes", name, note=note)

    def no(self, name, *, id=None, note=""):
        return self.vote("no", name, note=note)

    def set_count(self, option, value):
        for co in self.counts:
            if co["option"] == option:
                co["value"] = value
                break
        else:
            self.counts.append({"option": option, "value": value})


class OrderVoteEvent:
    """ A functor for applying order to voteEvents.
        A single OrderVoteEvent instance should be used for all bills in a scrape.
        The vote events of each bill must be processed in chronological order,
            but the processing of bills may be interleaved (needed in e.g. NH).
        Currently, it only fudges midnight dates (start_date and end_date)
            by adding the event sequence number in seconds
            to the start_date and end_date (if they are well-formed string dates)
        In the future, when there is an 'order' field on voteEvents,
            it should fill that as well.
        This fails softly and silently;
            if a valid string date is not found in start_date or end_date, the date is not touched.
        This assumes that times are reported as local time, not UTC.
            A UTC time that is local midnight will not be touched.
        Sometimes one chamber reports the time of a vote,
            but the other chamber reports only the date.  This is handled.
        See the unit tests for examples and more behavior.
    """

    _midnight = r"\d\d\d\d-\d\d-\d\dT00:00:00.*"
    _timeless = r"\d\d\d\d-\d\d-\d\d"

    class OrderBillVoteEvent:
        """ Order VoteEvents for a single bill
        """

        def __init__(self):
            self.order = 0  # voteEvent sequence number. 1st voteEvent is 1.

        def __call__(self, voteEvent):

            self.order += 1
            voteEvent.start_date = self._adjust_date(voteEvent.start_date)
            if hasattr(voteEvent, "end_date"):
                voteEvent.end_date = self._adjust_date(voteEvent.end_date)

        def _adjust_date(self, date):

            if not isinstance(date, str):
                return date

            if re.fullmatch(OrderVoteEvent._timeless, date):
                d2 = date + "T00:00:00"
            elif re.fullmatch(OrderVoteEvent._midnight, date):
                d2 = date
            else:
                return date

            assert self.order <= 60 * 60
            mins = "{:02d}".format(self.order // 60)
            secs = "{:02d}".format(self.order % 60)

            # yyyy-mm-ddThh:mm:dd+05:00
            # 0123456789012345678
            return d2[:14] + mins + ":" + secs + d2[19:]

    def __init__(self):
        self.orderers = {}

    def __call__(self, session_id, bill_id, voteEvent):
        """
        Record order of voteEvent within bill.

        The "order" field is not yet implemented; this fudges voteEvent start_date and end_date.
        See OrderVoteEvent docstring for details.

        :param session_id: session id
        :param bill_id: an identifier for the vote's bill
            that is at least unique within the session.
        :param voteEvent:
        :return: None
        """
        bill_orderer = self.orderers.get((session_id, bill_id))

        if not bill_orderer:
            bill_orderer = self.OrderBillVoteEvent()
            self.orderers[(session_id, bill_id)] = bill_orderer

        bill_orderer(voteEvent)



import pytest
from openstates.scrape import Post as ScrapePost
from openstates.importers import PostImporter, OrganizationImporter
from openstates.data.models import Organization, Post, Division, Jurisdiction
import datetime


def create_jurisdictions():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    Division.objects.create(id="ocd-division/country:us/state:nc", name="NC")
    Jurisdiction.objects.create(id="us", division_id="ocd-division/country:us")
    Jurisdiction.objects.create(id="nc", division_id="ocd-division/country:us/state:nc")


@pytest.mark.django_db
def test_full_post():
    create_jurisdictions()
    org = Organization.objects.create(
        name="United States Executive Branch",
        classification="executive",
        jurisdiction_id="us",
    )
    post = ScrapePost(
        label="executive",
        role="President",
        organization_id='~{"classification": "executive"}',
        start_date=datetime.date(2015, 5, 18),
        end_date="2015-05-19",
        maximum_memberships=2,
    )
    post.add_contact_detail(type="phone", value="555-555-1234", note="this is fake")
    post.add_link("http://example.com/link")

    # import post
    oi = OrganizationImporter("us")
    PostImporter("jurisdiction-id", oi).import_data([post.as_dict()])
    print(post.as_dict())

    # get person from db and assert it imported correctly
    p = Post.objects.get()
    assert "ocd-post" in p.id
    assert p.label == post.label
    assert p.role == post.role
    assert p.organization_id == org.id
    assert p.maximum_memberships == 2

    assert p.contact_details.all()[0].type == "phone"
    assert p.contact_details.all()[0].value == "555-555-1234"
    assert p.contact_details.all()[0].note == "this is fake"

    assert p.links.all()[0].url == "http://example.com/link"

    assert p.start_date == "2015-05-18"
    assert p.end_date == "2015-05-19"


@pytest.mark.django_db
def test_deduplication():
    create_jurisdictions()
    Organization.objects.create(
        id="us",
        name="United States Executive Branch",
        classification="executive",
        jurisdiction_id="us",
    )
    Organization.objects.create(
        id="nc",
        name="North Carolina Executive Branch",
        classification="executive",
        jurisdiction_id="nc",
    )
    pres = ScrapePost(
        label="executive",
        role="President",
        organization_id='~{"classification": "executive"}',
    )
    vp = ScrapePost(
        label="vice-executive",
        role="Vice President",
        organization_id='~{"classification": "executive"}',
    )
    gov = ScrapePost(
        label="executive",
        role="Governor",
        organization_id='~{"classification": "executive"}',
    )

    # ensure pres, vp and gov are all imported
    #   pres & gov - same label, different jurisdiction
    #   vp & pres - same jurisdiction, different label
    us_oi = OrganizationImporter("us")
    nc_oi = OrganizationImporter("nc")
    PostImporter("us", us_oi).import_data([pres.as_dict(), vp.as_dict()])
    PostImporter("nc", nc_oi).import_data([gov.as_dict()])
    assert Post.objects.count() == 3


@pytest.mark.django_db
def test_resolve_special_json_id():
    create_jurisdictions()
    Organization.objects.create(
        id="us",
        name="United States Executive Branch",
        classification="executive",
        jurisdiction_id="us",
    )
    Organization.objects.create(
        id="nc",
        name="North Carolina Executive Branch",
        classification="executive",
        jurisdiction_id="nc",
    )
    Post.objects.create(
        id="pres", label="executive", role="President", organization_id="us"
    )
    Post.objects.create(
        id="vpres", label="vice-executive", role="Vice President", organization_id="us"
    )
    Post.objects.create(
        id="gov", label="executive", role="Governor", organization_id="nc"
    )

    oi = OrganizationImporter("")
    assert PostImporter("us", oi).resolve_json_id('~{"label": "executive"}') == "pres"
    assert (
        PostImporter("us", oi).resolve_json_id('~{"label": "vice-executive"}')
        == "vpres"
    )
    assert PostImporter("nc", oi).resolve_json_id('~{"label": "executive"}') == "gov"



from collections import defaultdict
from itertools import chain


class CyclicGraphError(ValueError):
    """
    This exception is raised if the graph is Cyclic (or rather, when the
    sorting algorithm *knows* that the graph is Cyclic by hitting a snag
    in the top-sort)
    """

    pass


class Network(object):
    """
    This object (the `Network` object) handles keeping track of all the
    graph's nodes, and links between the nodes.

    The `Network' object is mostly used to topologically sort the nodes,
    to handle dependency resolution.
    """

    def __init__(self):
        self.nodes = set()
        self.edges = defaultdict(set)

    def add_node(self, node):
        """ Add a node to the graph (with no edges) """
        self.nodes.add(node)

    def add_edge(self, fro, to):
        """
        When doing topological sorting, the semantics of the edge mean that
        the depedency runs from the parent to the child - which is to say that
        the parent is required to be sorted *before* the child.

                  [ FROM ] ------> [ TO ]
        Committee on Finance -> Subcommittee of the Finance Committee on Budget
                             -> Subcommittee of the Finance Committee on Roads
        """
        self.add_node(fro)
        self.add_node(to)
        self.edges[fro].add(to)

    def leaf_nodes(self):
        """
        Return an interable of nodes with no edges pointing at them. This is
        helpful to find all nodes without dependencies.
        """
        # Now contains all nodes that contain dependencies.
        deps = {item for sublist in self.edges.values() for item in sublist}
        # contains all nodes *without* any dependencies (leaf nodes)
        return self.nodes - deps

    def prune_node(self, node, remove_backrefs=False):
        """
        remove node `node` from the network (including any edges that may
        have been pointing at `node`).
        """
        if not remove_backrefs:
            for fro, connections in self.edges.items():
                if node in self.edges[fro]:
                    raise ValueError(
                        """Attempting to remove a node with
                                     backrefs. You may consider setting
                                     `remove_backrefs` to true."""
                    )

        # OK. Otherwise, let's do our removal.

        self.nodes.remove(node)
        if node in self.edges:
            # Remove add edges from this node if we're pruning it.
            self.edges.pop(node)

        for fro, connections in self.edges.items():
            # Remove any links to this node (if they exist)
            if node in self.edges[fro]:
                # If we should remove backrefs:
                self.edges[fro].remove(node)

    def sort(self):
        """
        Return an iterable of nodes, toplogically sorted to correctly import
        dependencies before leaf nodes.
        """
        while self.nodes:
            iterated = False
            for node in self.leaf_nodes():
                iterated = True
                self.prune_node(node)
                yield node
            if not iterated:
                raise CyclicGraphError("Sorting has found a cyclic graph.")

    def dot(self):
        """
        Return a buffer that represents something dot(1) can render.
        """
        buff = "digraph graphname {"
        for fro in self.edges:
            for to in self.edges[fro]:
                buff += "%s -> %s;" % (fro, to)
        buff += "}"
        return buff

    def cycles(self):
        """
        Fairly expensive cycle detection algorithm. This method
        will return the shortest unique cycles that were detected.

        Debug usage may look something like:

        print("The following cycles were found:")
        for cycle in network.cycles():
            print("    ", " -> ".join(cycle))
        """

        def walk_node(node, seen):
            """
            Walk each top-level node we know about, and recurse
            along the graph.
            """
            if node in seen:
                yield (node,)
                return
            seen.add(node)
            for edge in self.edges[node]:
                for cycle in walk_node(edge, set(seen)):
                    yield (node,) + cycle

        # First, let's get a iterable of all known cycles.
        cycles = chain.from_iterable((walk_node(node, set()) for node in self.nodes))

        shortest = set()
        # Now, let's go through and sift through the cycles, finding
        # the shortest unique cycle known, ignoring cycles which contain
        # already known cycles.
        for cycle in sorted(cycles, key=len):
            for el in shortest:
                if set(el).issubset(set(cycle)):
                    break
            else:
                shortest.add(cycle)
        # And return that unique list.
        return shortest



from collections import defaultdict
from django.db.models import Q
from .base import BaseImporter
from ..exceptions import SameNameError
from ..data.models import (
    Person,
    PersonIdentifier,
    PersonName,
    PersonContactDetail,
    PersonLink,
    PersonSource,
)


class PersonImporter(BaseImporter):
    _type = "person"
    model_class = Person
    related_models = {
        "identifiers": (PersonIdentifier, "person_id", {}),
        "other_names": (PersonName, "person_id", {}),
        "contact_details": (PersonContactDetail, "person_id", {}),
        "links": (PersonLink, "person_id", {}),
        "sources": (PersonSource, "person_id", {}),
    }

    def _prepare_imports(self, dicts):
        dicts = list(super(PersonImporter, self)._prepare_imports(dicts))

        by_name = defaultdict(list)
        for _, person in dicts:
            by_name[person["name"]].append(person)
            for other in person["other_names"]:
                by_name[other["name"]].append(person)

        # check for duplicates
        for name, people in by_name.items():
            if len(people) > 1:
                for person in people:
                    if person["birth_date"] == "":
                        raise SameNameError(name)

        return dicts

    def limit_spec(self, spec):
        """
        Whenever we do a Pseudo ID lookup from the database, we need to limit
        based on the memberships -> organization -> jurisdiction, so we scope
        the resolution.
        """
        if list(spec.keys()) == ["name"]:
            # if we're just resolving on name, include other names and family name
            name = spec["name"]
            return (Q(name=name) | Q(other_names__name=name) | Q(family_name=name)) & Q(
                memberships__organization__jurisdiction_id=self.jurisdiction_id
            )
        spec["memberships__organization__jurisdiction_id"] = self.jurisdiction_id
        return spec

    def get_object(self, person):
        all_names = [person["name"]] + [o["name"] for o in person["other_names"]]

        matches = list(
            self.model_class.objects.filter(
                Q(memberships__organization__jurisdiction_id=self.jurisdiction_id),
                (Q(name__in=all_names) | Q(other_names__name__in=all_names)),
            ).distinct("id")
        )

        matches_length = len(matches)
        if matches_length == 1 and not matches[0].birth_date:
            return matches[0]
        elif matches_length == 0:
            raise self.model_class.DoesNotExist(
                "No Person: {} in {}".format(all_names, self.jurisdiction_id)
            )
        else:
            # Try and match based on birth_date.
            if person["birth_date"]:
                for match in matches:
                    if (
                        person["birth_date"]
                        and match.birth_date == person["birth_date"]
                    ):
                        return match

                # If we got here, no match based on birth_date, a new person?
                raise self.model_class.DoesNotExist(
                    "No Person: {} in {} with birth_date {}".format(
                        all_names, self.jurisdiction_id, person["birth_date"]
                    )
                )

            raise SameNameError(person["name"])



from .base import BaseImporter
from ..data.models import Membership, MembershipContactDetail, MembershipLink
from ..utils import get_pseudo_id
from ..exceptions import NoMembershipsError


class MembershipImporter(BaseImporter):
    _type = "membership"
    model_class = Membership
    related_models = {
        "contact_details": (MembershipContactDetail, "membership_id", {}),
        "links": (MembershipLink, "membership_id", {}),
    }

    def __init__(self, jurisdiction_id, person_importer, org_importer, post_importer):
        super(MembershipImporter, self).__init__(jurisdiction_id)
        self.person_importer = person_importer
        self.org_importer = org_importer
        self.post_importer = post_importer
        self.seen_person_ids = set()

    def get_object(self, membership):
        spec = {
            "organization_id": membership["organization_id"],
            "person_id": membership["person_id"],
            "label": membership["label"],
            "role": membership["role"],
            # if this is a historical role, only update historical roles
            "end_date": membership["end_date"],
        }

        # post_id is optional - might exist in DB but not scraped here?
        if membership["post_id"]:
            spec["post_id"] = membership["post_id"]

        if membership["person_name"]:
            spec["person_name"] = membership["person_name"]

        return self.model_class.objects.get(**spec)

    def prepare_for_db(self, data):
        # check if the organization is not tied to a jurisdiction
        if data["organization_id"].startswith("~"):
            pseudo_id = get_pseudo_id(data["organization_id"])
            is_party = pseudo_id.get("classification") == "party"
        else:
            # we have to assume it is not a party if we want to avoid doing a lookup here
            is_party = False

        data["organization_id"] = self.org_importer.resolve_json_id(
            data["organization_id"]
        )
        data["person_id"] = self.person_importer.resolve_json_id(
            data["person_id"], allow_no_match=True
        )
        data["post_id"] = self.post_importer.resolve_json_id(data["post_id"])
        if not is_party:
            # track that we had a membership for this person
            self.seen_person_ids.add(data["person_id"])
        return data

    def postimport(self):
        person_ids = (
            set(self.person_importer.json_to_db_id.values()) - self.seen_person_ids
        )
        if person_ids:
            reverse_id_dict = {
                v: k for k, v in self.person_importer.json_to_db_id.items()
            }
            person_ids = [reverse_id_dict[pid] for pid in person_ids]
            raise NoMembershipsError(person_ids)



import os
import json
import shutil
import tempfile
import pytest
from unittest import mock
from openstates.data.models import Person, Organization, Jurisdiction, Division
from openstates.scrape import Person as ScrapePerson
from openstates.scrape import Organization as ScrapeOrganization
from openstates.importers.base import omnihash, BaseImporter
from openstates.importers import PersonImporter, OrganizationImporter
from openstates.exceptions import UnresolvedIdError, DataImportError


def create_jurisdiction():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")


class FakeImporter(BaseImporter):
    _type = "test"


def test_omnihash_python_types():
    # string
    assert omnihash("test") == omnihash("test")
    # list
    assert omnihash(["this", "is", "a", "list"]) == omnihash(
        ["this", "is", "a", "list"]
    )
    # set
    assert omnihash({"and", "a", "set"}) == omnihash({"set", "set", "and", "a"})
    # dict w/ set and tuple as well
    assert omnihash({"a": {("fancy", "nested"): {"dict"}}}) == omnihash(
        {"a": {("fancy", "nested"): {"dict"}}}
    )


def test_import_directory():
    # write out some temp data to filesystem
    datadir = tempfile.mkdtemp()
    dicta = {"test": "A"}
    dictb = {"test": "B"}
    open(os.path.join(datadir, "test_a.json"), "w").write(json.dumps(dicta))
    open(os.path.join(datadir, "test_b.json"), "w").write(json.dumps(dictb))

    # simply ensure that import directory calls import_data with all dicts
    ti = FakeImporter("jurisdiction-id")
    with mock.patch.object(ti, attribute="import_data") as mockobj:
        ti.import_directory(datadir)

    # import_data should be called once
    assert mockobj.call_count == 1
    # kind of hacky, get the total list of args passed in
    arg_objs = list(mockobj.call_args[0][0])

    # 2 args only, make sure a and b are in there
    assert len(arg_objs) == 2
    assert dicta in arg_objs
    assert dictb in arg_objs

    # clean up datadir
    shutil.rmtree(datadir)


def test_apply_transformers():
    transformers = {
        "capitalize": lambda x: x.upper(),
        "cap_and_reverse": [lambda x: x.upper(), lambda y: y[::-1]],
        "never_used": lambda x: 1 / 0,
        "nested": {"replace": lambda x: "replaced"},
    }
    data = {
        "capitalize": "words",
        "cap_and_reverse": "simple",
        "nested": {"replace": None},
    }
    ti = FakeImporter("jid")
    ti.cached_transformers = transformers
    output = ti.apply_transformers(data)
    assert output["capitalize"] == "WORDS"
    assert output["cap_and_reverse"] == "ELPMIS"
    assert output["nested"]["replace"] == "replaced"


# doing these next few tests just on a Person because it is the same code that handles it
# but for completeness maybe it is better to do these on each type?


@pytest.mark.django_db
def test_deduplication_identical_object():
    p1 = ScrapePerson("Dwayne").as_dict()
    p2 = ScrapePerson("Dwayne").as_dict()
    PersonImporter("jid").import_data([p1, p2])

    assert Person.objects.count() == 1


@pytest.mark.django_db
def test_exception_on_identical_objects_in_import_stream():
    create_jurisdiction()
    # these two objects aren't identical, but refer to the same thing
    # at the moment we consider this an error (but there may be a better way to handle this?)
    o1 = ScrapeOrganization("X-Men", classification="unknown").as_dict()
    o2 = ScrapeOrganization(
        "X-Men", founding_date="1970", classification="unknown"
    ).as_dict()

    with pytest.raises(Exception):
        OrganizationImporter("jid").import_data([o1, o2])


@pytest.mark.django_db
def test_resolve_json_id():
    p1 = ScrapePerson("Dwayne").as_dict()
    p2 = ScrapePerson("Dwayne").as_dict()
    pi = PersonImporter("jid")

    # do import and get database id
    p1_id = p1["_id"]
    p2_id = p2["_id"]
    pi.import_data([p1, p2])
    db_id = Person.objects.get().id

    # simplest case
    assert pi.resolve_json_id(p1_id) == db_id
    # duplicate should resolve to same id
    assert pi.resolve_json_id(p2_id) == db_id
    # a null id should map to None
    assert pi.resolve_json_id(None) is None
    # no such id
    with pytest.raises(UnresolvedIdError):
        pi.resolve_json_id("this-is-invalid")


@pytest.mark.django_db
def test_invalid_fields():
    p1 = ScrapePerson("Dwayne").as_dict()
    p1["newfield"] = "shouldn't happen"

    with pytest.raises(DataImportError):
        PersonImporter("jid").import_data([p1])


@pytest.mark.django_db
def test_invalid_fields_related_item():
    p1 = ScrapePerson("Dwayne")
    p1.add_link("http://example.com")
    p1 = p1.as_dict()
    p1["links"][0]["test"] = 3

    with pytest.raises(DataImportError):
        PersonImporter("jid").import_data([p1])


@pytest.mark.django_db
def test_locked_field():
    create_jurisdiction()
    org = ScrapeOrganization("SHIELD").as_dict()
    oi = OrganizationImporter("jid")
    oi.import_data([org])

    # set date and lock field
    o = Organization.objects.get()
    o.dissolution_date = "2015"
    o.locked_fields = ["dissolution_date"]
    o.save()

    # reimport
    org = ScrapeOrganization("SHIELD").as_dict()
    oi = OrganizationImporter("jid")
    oi.import_data([org])

    o = Organization.objects.get()
    assert o.dissolution_date == "2015"
    assert o.locked_fields == ["dissolution_date"]

    # do it a third time to check for the locked_fields reversion issue
    org = ScrapeOrganization("SHIELD").as_dict()
    oi = OrganizationImporter("jid")
    oi.import_data([org])

    o = Organization.objects.get()
    assert o.dissolution_date == "2015"
    assert o.locked_fields == ["dissolution_date"]


@pytest.mark.django_db
def test_locked_field_subitem():
    create_jurisdiction()
    org = ScrapeOrganization("SHIELD")
    org.add_name("S.H.I.E.L.D.")
    oi = OrganizationImporter("jid")
    oi.import_data([org.as_dict()])

    # lock the field
    o = Organization.objects.get()
    o.locked_fields = ["other_names"]
    o.save()

    # reimport
    org = ScrapeOrganization("SHIELD").as_dict()
    oi = OrganizationImporter("jid")
    oi.import_data([org])

    o = Organization.objects.get()
    assert o.other_names.get().name == "S.H.I.E.L.D."



from django.db import models
from django.contrib.postgres.fields import ArrayField

from ..common import JURISDICTION_CLASSIFICATION_CHOICES
from .base import OCDBase, OCDIDField
from .division import Division


class Jurisdiction(OCDBase):
    """
    A Jurisdiction represents a logical unit of governance.

    Examples would include: the United States Federal Government, the Government
    of the District of Columbia, the Lexington-Fayette Urban County Government,
    or the Wake County Public School System.
    """

    id = OCDIDField(ocd_type="jurisdiction")
    name = models.CharField(
        max_length=300,
        help_text="The common name of the Jurisdiction, such as 'Wyoming.'",
    )
    url = models.URLField(
        max_length=2000, help_text="The primary website of the Jurisdiction."
    )
    classification = models.CharField(
        max_length=50,
        choices=JURISDICTION_CLASSIFICATION_CHOICES,
        default="government",
        db_index=True,
        help_text="The type of Jurisdiction being defined.",
    )
    feature_flags = ArrayField(
        base_field=models.TextField(),
        blank=True,
        default=list,
        help_text="A list of features that are present for data in this jurisdiction.",
    )
    division = models.ForeignKey(
        Division,
        related_name="jurisdictions",
        db_index=True,
        help_text="A link to a Division related to this Jurisdiction.",
        # don't allow deletion of a division that a Jurisdiction depends upon
        on_delete=models.PROTECT,
    )

    class Meta:
        db_table = "opencivicdata_jurisdiction"

    def __str__(self):
        return self.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import re
import io
import csv
from urllib.request import urlopen

PWD = os.path.abspath(os.path.dirname(__file__))
OCD_REMOTE_URL = (
    "https://raw.githubusercontent.com/opencivicdata/ocd-division-ids/master/"
    "identifiers/country-{}.csv"
)


class Division(object):
    _cache = {}

    @classmethod
    def all(self, country, from_csv=None):
        file_handle = None

        # Load from CSV if `from_csv` or `OCD_DIVISION_CSV` are set.
        if from_csv or "OCD_DIVISION_CSV" in os.environ:
            if not from_csv:
                from_csv = os.environ.get("OCD_DIVISION_CSV").format(country)
            try:
                file_handle = io.open(from_csv, encoding="utf8")
            except FileNotFoundError:
                raise ValueError("Couldn't open CSV file {}".format(from_csv))

        # Load from URL otherwise.
        if not file_handle:
            file_handle = io.StringIO(
                urlopen(OCD_REMOTE_URL.format(country)).read().decode("utf-8")
            )

        for row in csv.DictReader(file_handle):
            yield Division(**row)

    @classmethod
    def get(self, division, from_csv=None):
        if division not in self._cache:
            # Figure out the country.
            if not re.match(r"ocd-division/country:\w{2}", division):
                raise ValueError("Invalid OCD format.")
            country = re.findall(r"country:(\w{2})", division)[0]

            # Load all divisions into cache.
            for d in self.all(country, from_csv):
                pass

            if division not in self._cache:
                raise ValueError("Division not found: {}".format(division))

        return self._cache[division]

    def __init__(self, id, name, **kwargs):
        self._cache[id] = self
        self.id = id
        self.name = name
        self.sameAs = kwargs.pop("sameAs", None)
        valid_through = kwargs.pop("validThrough", None)
        if valid_through:
            self.valid_through = valid_through

        # set parent and _type
        parent, own_id = id.rsplit("/", 1)
        if parent == "ocd-division":
            self.parent = None
        else:
            self.parent = self._cache.get(parent)
            if self.parent:
                self.parent._children.append(self)
            else:
                # TODO: keep a list of unassigned parents for later reconciliation
                pass

        self._type = own_id.split(":")[0]

        # other attrs
        self.attrs = kwargs
        self.names = []
        self._children = []

    def children(self, _type=None, duplicates=True, levels=1):
        for d in self._children:
            if (not _type or d._type == _type) and (duplicates or not d.sameAs):
                yield d
                if levels > 1:
                    for c in d.children(_type, duplicates, levels - 1):
                        yield c

    def __str__(self):
        return "{} - {}".format(self.id, self.name)



#!/u/bin/env python
# -*- coding: utf-8 -*-
from openstates.data.divisions import Division


def test_get():
    div = Division.get("ocd-division/country:de/state:by/cd:248")
    assert div.name == "Bad Kissingen"
    assert div.name in str(div)


def test_children():
    us = Division.get("ocd-division/country:ua")
    assert len(list(us.children("region", duplicates=False))) == 25



import pytest
from django.core.management import call_command
from openstates.data.models import Division


@pytest.mark.django_db
def test_loaddivisions():
    assert Division.objects.count() == 0
    call_command("loaddivisions", "in")
    assert Division.objects.count() > 0
    assert Division.objects.count() == Division.objects.filter(country="in").count()
    assert (
        Division.objects.filter(country="in", subtype1="state", subtype2="").count()
        == 29
    )

    # Include a (very weak) check for idempotency
    call_command("loaddivisions", "in")
    assert (
        Division.objects.filter(country="in", subtype1="state", subtype2="").count()
        == 29
    )



from ..data.models import Post, PostContactDetail, PostLink
from .base import BaseImporter


class PostImporter(BaseImporter):
    _type = "post"
    model_class = Post
    related_models = {
        "contact_details": (PostContactDetail, "post_id", {}),
        "links": (PostLink, "post_id", {}),
    }

    def __init__(self, jurisdiction_id, org_importer):
        super(PostImporter, self).__init__(jurisdiction_id)
        self.org_importer = org_importer

    def prepare_for_db(self, data):
        data["organization_id"] = self.org_importer.resolve_json_id(
            data["organization_id"]
        )
        return data

    def get_object(self, post):
        spec = {"organization_id": post["organization_id"], "label": post["label"]}

        if post["role"]:
            spec["role"] = post["role"]

        return self.model_class.objects.get(**spec)

    def limit_spec(self, spec):
        spec["organization__jurisdiction_id"] = self.jurisdiction_id
        return spec



import re
import pytest
from openstates.scrape import Bill as ScrapeBill
from openstates.scrape import Person as ScrapePerson
from openstates.scrape import Organization as ScrapeOrganization
from openstates.importers import BillImporter, OrganizationImporter, PersonImporter
from openstates.data.models import (
    Jurisdiction,
    Person,
    Organization,
    Membership,
    Division,
    Bill,
)


def create_jurisdiction():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    j = Jurisdiction.objects.create(id="jid", division_id="ocd-division/country:us")
    j.legislative_sessions.create(identifier="1899", name="1899")
    j.legislative_sessions.create(identifier="1900", name="1900")


def create_org():
    return Organization.objects.create(
        id="org-id", name="House", classification="lower", jurisdiction_id="jid"
    )


@pytest.mark.django_db
def test_full_bill():
    create_jurisdiction()
    sp = ScrapePerson("Adam Smith")
    org = ScrapeOrganization(name="House", classification="lower")
    com = ScrapeOrganization(
        name="Arbitrary Committee", classification="committee", parent_id=org._id
    )

    oldbill = ScrapeBill(
        "HB 99",
        "1899",
        "Axe & Tack Tax Act",
        classification="tax bill",
        from_organization=org._id,
    )

    bill = ScrapeBill(
        "HB 1",
        "1900",
        "Axe & Tack Tax Act",
        classification="tax bill",
        from_organization=org._id,
    )
    bill.subject = ["taxes", "axes"]
    bill.add_identifier("SB 9")
    bill.add_title("Tack & Axe Tax Act")
    bill.add_action("introduced in house", "1900-04-01", chamber="lower")
    act = bill.add_action("sent to arbitrary committee", "1900-04-04", chamber="lower")
    act.add_related_entity("arbitrary committee", "organization", com._id)
    bill.add_related_bill(
        "HB 99", legislative_session="1899", relation_type="prior-session"
    )
    bill.add_sponsorship(
        "Adam Smith",
        classification="extra sponsor",
        entity_type="person",
        primary=False,
        entity_id=sp._id,
    )
    bill.add_sponsorship(
        "Jane Smith", classification="lead sponsor", entity_type="person", primary=True
    )
    bill.add_abstract(
        "This is an act about axes and taxes and tacks.",
        note="official",
        date="1969-10-20",
    )
    bill.add_document_link(
        "Fiscal Note", "http://example.com/fn.pdf", media_type="application/pdf"
    )
    bill.add_document_link(
        "Fiscal Note", "http://example.com/fn.html", media_type="text/html"
    )
    bill.add_version_link(
        "Fiscal Note", "http://example.com/v/1", media_type="text/html"
    )
    bill.add_source("http://example.com/source")

    # import bill
    oi = OrganizationImporter("jid")
    oi.import_data([org.as_dict(), com.as_dict()])

    pi = PersonImporter("jid")
    pi.import_data([sp.as_dict()])

    BillImporter("jid", oi, pi).import_data([oldbill.as_dict(), bill.as_dict()])

    # get bill from db and assert it imported correctly
    b = Bill.objects.get(identifier="HB 1")
    assert b.from_organization.classification == "lower"
    assert b.identifier == bill.identifier
    assert b.title == bill.title
    assert b.classification == bill.classification
    assert b.subject == ["taxes", "axes"]
    assert b.abstracts.get().note == "official"
    assert b.abstracts.get().date == "1969-10-20"

    # other_title, other_identifier added
    assert b.other_titles.get().title == "Tack & Axe Tax Act"
    assert b.other_identifiers.get().identifier == "SB 9"

    # actions
    actions = list(b.actions.all())
    assert len(actions) == 2
    # ensure order was preserved (if this breaks it'll be intermittent)
    assert actions[0].organization == Organization.objects.get(classification="lower")
    assert actions[0].description == "introduced in house"
    assert actions[1].description == "sent to arbitrary committee"
    assert actions[1].related_entities.get().organization == Organization.objects.get(
        classification="committee"
    )

    # related_bills were added
    rb = b.related_bills.get()
    assert rb.identifier == "HB 99"

    # and bill got resolved
    assert rb.related_bill.identifier == "HB 99"

    # sponsors added, linked & unlinked
    sponsorships = b.sponsorships.all()
    assert len(sponsorships) == 2
    person = Person.objects.get(name="Adam Smith")
    for ss in sponsorships:
        if ss.primary:
            assert ss.person is None
            assert ss.organization is None
        else:
            assert ss.person == person

    # versions & documents with their links
    versions = b.versions.all()
    assert len(versions) == 1
    assert versions[0].links.count() == 1
    documents = b.documents.all()
    assert len(documents) == 1
    assert documents[0].links.count() == 2

    # sources
    assert b.sources.count() == 1


@pytest.mark.django_db
def test_bill_chamber_param():
    create_jurisdiction()
    org = create_org()

    bill = ScrapeBill(
        "HB 1", "1900", "Axe & Tack Tax Act", classification="tax bill", chamber="lower"
    )

    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")
    BillImporter("jid", oi, pi).import_data([bill.as_dict()])

    assert Bill.objects.get().from_organization_id == org.id


@pytest.mark.django_db
def test_bill_update():
    create_jurisdiction()
    create_org()

    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")

    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")
    _, what = BillImporter("jid", oi, pi).import_item(bill.as_dict())
    assert what == "insert"
    _, what = BillImporter("jid", oi, pi).import_item(bill.as_dict())
    assert what == "noop"

    # ensure no new object was created
    assert Bill.objects.count() == 1

    # test basic update
    bill = ScrapeBill("HB 1", "1900", "1st Bill", chamber="lower")
    _, what = BillImporter("jid", oi, pi).import_item(bill.as_dict())
    assert what == "update"
    assert Bill.objects.get().title == "1st Bill"


@pytest.mark.django_db
def test_bill_update_because_of_subitem():
    create_jurisdiction()
    create_org()
    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")

    # initial bill
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_action("this is an action", chamber="lower", date="1900-01-01")
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["insert"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 1
    last_updated = obj.updated_at

    # now let's make sure we get updated when there are second actions
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_action("this is an action", chamber="lower", date="1900-01-01")
    bill.add_action("this is a second action", chamber="lower", date="1900-01-02")
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 2
    assert obj.updated_at > last_updated

    # same 2 actions, noop
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_action("this is an action", chamber="lower", date="1900-01-01")
    bill.add_action("this is a second action", chamber="lower", date="1900-01-02")
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["noop"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 2

    # same 2 actions, different order, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_action("this is a second action", chamber="lower", date="1900-01-02")
    bill.add_action("this is an action", chamber="lower", date="1900-01-01")
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 2

    # different 2 actions, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_action("this is an action", chamber="lower", date="1900-01-01")
    bill.add_action(
        "this is a different second action", chamber="lower", date="1900-01-02"
    )
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 2

    # delete an action, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_action("this is a second action", chamber="lower", date="1900-01-02")
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 1

    # delete all actions, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 0

    # and back to initial status, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_action("this is an action", chamber="lower", date="1900-01-01")
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.actions.count() == 1


@pytest.mark.django_db
def test_bill_update_subsubitem():
    create_jurisdiction()
    create_org()
    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")

    # initial sub-subitem
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_version_link(
        "printing", "http://example.com/test.pdf", media_type="application/pdf"
    )
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["insert"] == 1
    obj = Bill.objects.get()
    assert obj.versions.count() == 1
    assert obj.versions.get().links.count() == 1

    # a second subsubitem, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_version_link(
        "printing", "http://example.com/test.pdf", media_type="application/pdf"
    )
    bill.add_version_link(
        "printing", "http://example.com/test.text", media_type="text/plain"
    )
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.versions.count() == 1
    assert obj.versions.get().links.count() == 2

    # same thing, noop
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_version_link(
        "printing", "http://example.com/test.pdf", media_type="application/pdf"
    )
    bill.add_version_link(
        "printing", "http://example.com/test.text", media_type="text/plain"
    )
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["noop"] == 1
    obj = Bill.objects.get()
    assert obj.versions.count() == 1
    assert obj.versions.get().links.count() == 2

    # different link for second one, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_version_link(
        "printing", "http://example.com/test.pdf", media_type="application/pdf"
    )
    bill.add_version_link(
        "printing", "http://example.com/diff-link.txt", media_type="text/plain"
    )
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.versions.count() == 1
    assert obj.versions.get().links.count() == 2

    # delete one, update
    bill = ScrapeBill("HB 1", "1900", "First Bill", chamber="lower")
    bill.add_version_link(
        "printing", "http://example.com/test.pdf", media_type="application/pdf"
    )
    result = BillImporter("jid", oi, pi).import_data([bill.as_dict()])
    assert result["bill"]["update"] == 1
    obj = Bill.objects.get()
    assert obj.versions.count() == 1
    assert obj.versions.get().links.count() == 1


@pytest.mark.django_db
def test_bill_sponsor_by_identifier():
    create_jurisdiction()
    org = create_org()

    bill = ScrapeBill(
        "HB 1", "1900", "Axe & Tack Tax Act", classification="tax bill", chamber="lower"
    )
    bill.add_sponsorship_by_identifier(
        name="SNODGRASS",
        classification="sponsor",
        entity_type="person",
        primary=True,
        identifier="TOTALLY_REAL_ID",
        scheme="TOTALLY_REAL_SCHEME",
    )

    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")

    zs = ScrapePerson(name="Zadock Snodgrass")
    zs.add_identifier(identifier="TOTALLY_REAL_ID", scheme="TOTALLY_REAL_SCHEME")
    pi.import_data([zs.as_dict()])
    za_db = Person.objects.get()
    Membership.objects.create(person_id=za_db.id, organization_id=org.id)

    BillImporter("jid", oi, pi).import_data([bill.as_dict()])

    obj = Bill.objects.get()
    (entry,) = obj.sponsorships.all()
    assert entry.person.name == "Zadock Snodgrass"


@pytest.mark.django_db
def test_bill_sponsor_limit_lookup():
    create_jurisdiction()
    org = create_org()

    bill = ScrapeBill(
        "HB 1", "1900", "Axe & Tack Tax Act", classification="tax bill", chamber="lower"
    )
    bill.add_sponsorship_by_identifier(
        name="SNODGRASS",
        classification="sponsor",
        entity_type="person",
        primary=True,
        identifier="TOTALLY_REAL_ID",
        scheme="TOTALLY_REAL_SCHEME",
    )

    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")

    zs = ScrapePerson(name="Zadock Snodgrass", birth_date="1800-01-01")
    zs.add_identifier(identifier="TOTALLY_REAL_ID", scheme="TOTALLY_REAL_SCHEME")
    pi.import_data([zs.as_dict()])

    za_db = Person.objects.get()
    Membership.objects.create(person_id=za_db.id, organization_id=org.id)

    zs2 = ScrapePerson(name="Zadock Snodgrass", birth_date="1900-01-01")
    zs2.add_identifier(identifier="TOTALLY_REAL_ID", scheme="TOTALLY_REAL_SCHEME")

    # This is contrived and perhaps broken, but we're going to check this.
    # We *really* don't want to *ever* cross jurisdiction bounds.
    PersonImporter("another-jurisdiction").import_data([zs.as_dict()])

    BillImporter("jid", oi, pi).import_data([bill.as_dict()])

    obj = Bill.objects.get()
    (entry,) = obj.sponsorships.all()
    assert entry.person.name == "Zadock Snodgrass"
    assert entry.person.birth_date == "1800-01-01"


@pytest.mark.django_db
def test_bill_action_extras():
    create_jurisdiction()
    create_org()

    bill = ScrapeBill(
        "HB 1", "1900", "Axe & Tack Tax Act", classification="tax bill", chamber="lower"
    )
    bill.add_action("sample", "1900-01-01", chamber="lower", extras={"test": 3})

    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")

    BillImporter("jid", oi, pi).import_data([bill.as_dict()])

    b = Bill.objects.get()
    assert b.actions.all()[0].extras == {"test": 3}


@pytest.mark.django_db
def test_fix_bill_id():
    create_jurisdiction()
    create_org()

    bill = ScrapeBill(
        "HB1", "1900", "Test Bill ID", classification="bill", chamber="lower"
    )

    oi = OrganizationImporter("jid")
    pi = PersonImporter("jid")

    from openstates.settings import IMPORT_TRANSFORMERS

    IMPORT_TRANSFORMERS["bill"] = {
        "identifier": lambda x: re.sub(r"([A-Z]*)\s*0*([-\d]+)", r"\1 \2", x, 1)
    }
    bi = BillImporter("jid", oi, pi)
    bi.import_data([bill.as_dict()])
    IMPORT_TRANSFORMERS["bill"] = {}

    b = Bill.objects.get()
    assert b.identifier == "HB 1"



import pytest
from openstates.utils.topsort import Network, CyclicGraphError


def chash(cycles):
    """
    Hash a cycle, useful for comparing sets of cycles.

    This checks the sorted set of each of the nodes in the cycle. This
    is *not* a perfect check, but it's useful so that we can create a set
    of these hashes, and check that they all match.

    It's not perfect, since D -> A -> B will be the same as B -> A -> D,
    but since this is only used in the testing logic, we can ensure
    that we handle it correctly in the testcases.

    (Implicit warning: Don't use this anywhere important.)
    """
    return {"".join(sorted(set(x))) for x in cycles}


def test_sort_order_basic():
    network = Network()
    network.add_node("A")
    network.add_node("B")
    network.add_node("C")

    network.add_edge("A", "B")
    network.add_edge("B", "C")

    assert (list(network.sort())) == ["A", "B", "C"]


def test_sort_order_double():
    network = Network()
    network.add_node("A")
    network.add_node("B")
    network.add_node("C")

    network.add_edge("A", "B")
    network.add_edge("A", "C")
    network.add_edge("C", "B")

    # A  =>  B
    #       /
    # A => C

    assert (list(network.sort())) == ["A", "C", "B"]


def test_sort_order_staged():
    network = Network()

    network.add_node("A1")
    network.add_node("A2")
    network.add_node("A3")

    network.add_edge("A1", "A2")
    network.add_edge("A1", "A3")
    network.add_edge("A2", "A3")

    network.add_node("B1")
    network.add_node("B2")
    network.add_node("B3")

    network.add_edge("B1", "B2")
    network.add_edge("B1", "B3")
    network.add_edge("B2", "B3")

    network.add_edge("B1", "A1")

    network.add_node("C1")
    network.add_node("C2")
    network.add_node("C3")

    network.add_edge("C1", "C2")
    network.add_edge("C1", "C3")
    network.add_edge("C2", "C3")

    network.add_edge("C1", "A1")
    network.add_edge("C1", "B1")

    network.add_edge("C1", "B1")
    network.add_edge("B1", "A1")
    network.add_edge("A1", "C2")
    network.add_edge("A1", "C3")

    # with open("/home/tag/debug.dot", 'w') as fd:
    #     fd.write(network.dot())

    sorted_order = list(network.sort())

    assert sorted_order.pop(0) == "C1"
    assert sorted_order.pop(0) == "B1"
    assert sorted_order.pop(0) in ("A1", "B2")
    #                          ^^ This makes more sense after you dot debug it
    assert sorted_order.pop(0) in ("A1", "B2")


def test_cyclic_graph_error_simple():
    network = Network()
    network.add_node("A")
    network.add_node("B")
    network.add_edge("A", "B")
    network.add_edge("B", "A")

    with pytest.raises(CyclicGraphError):
        list(network.sort())


def test_cyclic_graph_error_indirect():
    network = Network()
    network.add_node("A")
    network.add_node("B")
    network.add_node("C")

    network.add_edge("A", "B")
    network.add_edge("B", "C")
    network.add_edge("C", "A")

    with pytest.raises(CyclicGraphError):
        list(network.sort())


def test_cyclic_graph_error_massive():
    network = Network()

    entries = ["A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "A"]
    for i, e in enumerate(entries[:-1]):
        network.add_node(e)
        network.add_edge(e, entries[1 + i])

    with pytest.raises(CyclicGraphError):
        list(network.sort())


def test_link_before_nodes():
    network = Network()

    network.add_edge("A", "B")
    network.add_edge("B", "C")
    network.add_edge("C", "D")

    network.add_node("A")
    network.add_node("B")
    network.add_node("C")
    network.add_node("D")

    assert list(network.sort()) == ["A", "B", "C", "D"]


def test_internal_node_removal():
    network = Network()

    network.add_node("A")
    network.add_node("B")
    network.add_node("C")
    network.add_node("D")

    network.add_edge("A", "B")
    network.add_edge("B", "C")
    network.add_edge("C", "D")
    network.add_edge("A", "C")  # Useful for ensuring the ending list
    # is deterministic.

    # Ensure that we can't remove an internal node without a ValueError
    # by default.
    with pytest.raises(ValueError):
        network.prune_node("B")

    # OK. Now that we know that works, let's prune it harder.
    network.prune_node("B", remove_backrefs=True)

    # And make sure "B" is gone.
    assert list(network.sort()) == ["A", "C", "D"]


def test_dot_debug():
    network = Network()

    network.add_node("A")
    network.add_node("B")
    network.add_edge("A", "B")

    dot = network.dot()
    assert dot == "digraph graphname {A -> B;}"


def test_cycles_simple():
    network = Network()
    network.add_node("A")
    network.add_node("B")
    network.add_edge("A", "B")
    network.add_edge("B", "A")
    assert chash(network.cycles()) == chash([("A", "B", "A")])


def test_cycles_complex():
    network = Network()
    network.add_node("A")
    network.add_node("B")
    network.add_node("C")
    network.add_node("D")

    network.add_edge("A", "B")
    network.add_edge("B", "C")
    network.add_edge("C", "D")
    network.add_edge("D", "A")

    network.add_edge("D", "C")
    network.add_edge("C", "B")
    network.add_edge("B", "D")

    # with open("/home/tag/debug.dot", 'w') as fd:
    #     fd.write(network.dot())

    assert chash(network.cycles()) == chash(
        [("B", "C", "B"), ("C", "D", "C"), ("A", "B", "D", "A")]
    )



import pytest
from openstates.data.models import Organization, Jurisdiction, Division
from openstates.scrape import Organization as ScrapeOrganization
from openstates.importers import OrganizationImporter
from openstates.exceptions import UnresolvedIdError, SameOrgNameError


def create_jurisdictions():
    Division.objects.create(id="ocd-division/country:us", name="USA")
    Jurisdiction.objects.create(id="jid1", division_id="ocd-division/country:us")
    Jurisdiction.objects.create(id="jid2", division_id="ocd-division/country:us")


def create_org():
    o = Organization.objects.create(
        name="United Nations", jurisdiction_id="jid1", classification="international"
    )
    o.other_names.create(name="UN")


@pytest.mark.django_db
def test_full_organization():
    create_jurisdictions()
    org = ScrapeOrganization("United Nations", classification="international")
    org.add_identifier("un")
    org.add_name("UN", start_date="1945")
    org.add_contact_detail(type="phone", value="555-555-1234", note="this is fake")
    org.add_link("http://example.com/link")
    org.add_source("http://example.com/source")

    # import org
    od = org.as_dict()
    OrganizationImporter("jid1").import_data([od])

    # get person from db and assert it imported correctly
    o = Organization.objects.get()
    assert "ocd-organization" in o.id
    assert o.name == org.name

    assert o.identifiers.all()[0].identifier == "un"
    assert o.identifiers.all()[0].scheme == ""

    assert o.other_names.all()[0].name == "UN"
    assert o.other_names.all()[0].start_date == "1945"

    assert o.contact_details.all()[0].type == "phone"
    assert o.contact_details.all()[0].value == "555-555-1234"
    assert o.contact_details.all()[0].note == "this is fake"

    assert o.links.all()[0].url == "http://example.com/link"
    assert o.sources.all()[0].url == "http://example.com/source"


@pytest.mark.django_db
def test_deduplication_similar_but_different():
    create_jurisdictions()
    o1 = ScrapeOrganization("United Nations", classification="international")
    # different classification
    o2 = ScrapeOrganization("United Nations", classification="global")
    # different name
    o3 = ScrapeOrganization("United Nations of Earth", classification="international")
    # has a parent
    o4 = ScrapeOrganization(
        "United Nations", classification="international", parent_id=o1._id
    )

    # similar, but no duplicates
    orgs = [o1.as_dict(), o2.as_dict(), o3.as_dict(), o4.as_dict()]
    OrganizationImporter("jid1").import_data(orgs)
    assert Organization.objects.count() == 4

    # should get a new one  when jurisdiction_id changes
    o5 = ScrapeOrganization("United Nations", classification="international")
    OrganizationImporter("jid2").import_data([o5.as_dict()])
    assert Organization.objects.count() == 5


@pytest.mark.django_db
def test_deduplication_other_name_exists():
    create_jurisdictions()
    create_org()
    org = ScrapeOrganization("UN", classification="international")
    od = org.as_dict()
    OrganizationImporter("jid1").import_data([od])
    assert Organization.objects.all().count() == 1


@pytest.mark.django_db
def test_deduplication_other_name_overlaps():
    create_jurisdictions()
    create_org()
    org = ScrapeOrganization("The United Nations", classification="international")
    org.add_name("United Nations")
    od = org.as_dict()
    OrganizationImporter("jid1").import_data([od])
    assert Organization.objects.all().count() == 1


@pytest.mark.django_db
def test_deduplication_error_overlaps():
    create_jurisdictions()

    Organization.objects.create(
        name="World Wrestling Federation",
        classification="international",
        jurisdiction_id="jid1",
    )
    wildlife = Organization.objects.create(
        name="World Wildlife Fund",
        classification="international",
        jurisdiction_id="jid1",
    )
    wildlife.other_names.create(name="WWF")

    org = ScrapeOrganization(
        "World Wrestling Federation", classification="international"
    )
    org.add_name("WWF")
    od = org.as_dict()
    with pytest.raises(SameOrgNameError):
        OrganizationImporter("jid1").import_data([od])


@pytest.mark.django_db
def test_deduplication_overlap_name_distinct_juris():
    create_jurisdictions()

    org_jid_1 = Organization.objects.create(
        name="World Wrestling Federation",
        classification="international",
        jurisdiction_id="jid1",
    )
    org_jid_1.other_names.create(name="WWF")

    org = ScrapeOrganization(name="WWF", classification="international")
    org.add_name("WWF")

    oi1 = OrganizationImporter("jid1")
    oi1.import_item(org.as_dict())
    assert Organization.objects.count() == 1

    oi2 = OrganizationImporter("jid2")
    oi2.import_item(org.as_dict())
    assert Organization.objects.count() == 2


@pytest.mark.django_db
def test_deduplication_parties():
    create_jurisdictions()
    party = ScrapeOrganization("Wild", classification="party")
    OrganizationImporter("jid1").import_data([party.as_dict()])
    assert Organization.objects.count() == 1

    # parties shouldn't get jurisdiction id attached, so don't differ on import
    party = ScrapeOrganization("Wild", classification="party")
    OrganizationImporter("jid2").import_data([party.as_dict()])
    assert Organization.objects.count() == 1


@pytest.mark.django_db
def test_deduplication_prevents_identical():
    create_jurisdictions()
    org1 = ScrapeOrganization("United Nations", classification="international")
    org2 = ScrapeOrganization(
        "United Nations", classification="international", founding_date="1945"
    )
    OrganizationImporter("jid1").import_data([org1.as_dict()])
    assert Organization.objects.count() == 1

    OrganizationImporter("jid1").import_data([org2.as_dict()])
    assert Organization.objects.count() == 1


@pytest.mark.django_db
def test_pseudo_ids():
    create_jurisdictions()
    wild = Organization.objects.create(id="1", name="Wild", classification="party")
    senate = Organization.objects.create(
        id="2", name="Senate", classification="upper", jurisdiction_id="jid1"
    )
    house = Organization.objects.create(
        id="3", name="House", classification="lower", jurisdiction_id="jid1"
    )
    un = Organization.objects.create(
        id="4",
        name="United Nations",
        classification="international",
        jurisdiction_id="jid2",
    )

    oi1 = OrganizationImporter("jid1")
    assert oi1.resolve_json_id('~{"classification":"upper"}') == senate.id
    assert oi1.resolve_json_id('~{"classification":"lower"}') == house.id
    assert oi1.resolve_json_id('~{"classification":"party", "name":"Wild"}') == wild.id

    with pytest.raises(UnresolvedIdError):
        oi1.resolve_json_id(
            '~{"classification":"international", "name":"United Nations"}'
        )

    oi2 = OrganizationImporter("jid2")
    assert (
        oi2.resolve_json_id(
            '~{"classification":"international", "name":"United Nations"}'
        )
        == un.id
    )


@pytest.mark.django_db
def test_parent_id_resolution():
    create_jurisdictions()
    parent = ScrapeOrganization("UN", classification="international")
    child = ScrapeOrganization("UNESCO", classification="unknown", parent_id=parent._id)
    OrganizationImporter("jid1").import_data([parent.as_dict(), child.as_dict()])
    assert Organization.objects.count() == 2
    assert Organization.objects.get(name="UN").children.count() == 1
    assert Organization.objects.get(name="UNESCO").parent.name == "UN"


@pytest.mark.django_db
def test_pseudo_parent_id_resolution():
    create_jurisdictions()
    parent = ScrapeOrganization("UN", classification="international")
    child = ScrapeOrganization(
        "UNESCO",
        classification="unknown",
        parent_id='~{"classification": "international"}',
    )
    OrganizationImporter("jid1").import_data([parent.as_dict(), child.as_dict()])
    assert Organization.objects.count() == 2
    assert Organization.objects.get(name="UN").children.count() == 1
    assert Organization.objects.get(name="UNESCO").parent.name == "UN"


@pytest.mark.django_db
def test_extras_organization():
    create_jurisdictions()
    org = ScrapeOrganization("United Nations", classification="international")
    org.extras = {"hello": "world", "foo": {"bar": "baz"}}
    od = org.as_dict()
    OrganizationImporter("jid1").import_data([od])
    o = Organization.objects.get()
    assert o.extras["foo"]["bar"] == "baz"



from django.db.models import Q
from .base import BaseImporter
from ..data.models import (
    Organization,
    OrganizationIdentifier,
    OrganizationName,
    OrganizationContactDetail,
    OrganizationLink,
    OrganizationSource,
)
from ..utils import get_pseudo_id
from ..utils.topsort import Network
from ..exceptions import UnresolvedIdError, InternalError, SameOrgNameError


class OrganizationImporter(BaseImporter):
    _type = "organization"
    model_class = Organization
    related_models = {
        "identifiers": (OrganizationIdentifier, "organization_id", {}),
        "other_names": (OrganizationName, "organization_id", {}),
        "contact_details": (OrganizationContactDetail, "organization_id", {}),
        "links": (OrganizationLink, "organization_id", {}),
        "sources": (OrganizationSource, "organization_id", {}),
    }

    def get_object(self, org):
        spec = {"classification": org["classification"], "parent_id": org["parent_id"]}

        # add jurisdiction_id unless this is a party
        jid = org.get("jurisdiction_id")
        if jid:
            spec["jurisdiction_id"] = jid

        all_names = [org["name"]] + [o["name"] for o in org["other_names"]]

        query = Q(**spec) & (Q(name__in=all_names) | Q(other_names__name__in=all_names))
        matches = list(self.model_class.objects.filter(query).distinct("id"))
        matches_length = len(matches)
        if matches_length == 1:
            return matches[0]
        elif matches_length == 0:
            raise self.model_class.DoesNotExist(
                "No Organization: {} in {}".format(all_names, self.jurisdiction_id)
            )
        else:
            raise SameOrgNameError(org["name"])

    def prepare_for_db(self, data):
        data["parent_id"] = self.resolve_json_id(data["parent_id"])

        if data["classification"] != "party":
            data["jurisdiction_id"] = self.jurisdiction_id
        return data

    def limit_spec(self, spec):
        if spec.get("classification") != "party":
            spec["jurisdiction_id"] = self.jurisdiction_id

        name = spec.pop("name", None)
        if name:
            return Q(**spec) & (Q(name=name) | Q(other_names__name=name))
        return spec

    def _prepare_imports(self, dicts):
        """ an override for prepare imports that sorts the imports by parent_id dependencies """
        # all pseudo parent ids we've seen
        pseudo_ids = set()
        # pseudo matches
        pseudo_matches = {}

        # get prepared imports from parent
        prepared = dict(super(OrganizationImporter, self)._prepare_imports(dicts))

        # collect parent pseudo_ids
        for _, data in prepared.items():
            parent_id = data.get("parent_id", None) or ""
            if parent_id.startswith("~"):
                pseudo_ids.add(parent_id)

        # turn pseudo_ids into a tuple of dictionaries
        pseudo_ids = [(ppid, get_pseudo_id(ppid)) for ppid in pseudo_ids]

        # loop over all data again, finding the pseudo ids true json id
        for json_id, data in prepared.items():
            # check if this matches one of our ppids
            for ppid, spec in pseudo_ids:
                match = True
                for k, v in spec.items():
                    if data[k] != v:
                        match = False
                        break
                if match:
                    if ppid in pseudo_matches:
                        raise UnresolvedIdError(
                            "multiple matches for pseudo id: " + ppid
                        )
                    pseudo_matches[ppid] = json_id

        # toposort the nodes so parents are imported first
        network = Network()
        in_network = set()
        import_order = []

        for json_id, data in prepared.items():
            parent_id = data.get("parent_id", None)

            # resolve pseudo_ids to their json id before building the network
            if parent_id in pseudo_matches:
                parent_id = pseudo_matches[parent_id]

            network.add_node(json_id)
            if parent_id:
                # Right. There's an import dep. We need to add the edge from
                # the parent to the current node, so that we import the parent
                # before the current node.
                network.add_edge(parent_id, json_id)

        # resolve the sorted import order
        for jid in network.sort():
            import_order.append((jid, prepared[jid]))
            in_network.add(jid)

        # ensure all data made it into network (paranoid check, should never fail)
        if in_network != set(prepared.keys()):  # pragma: no cover
            raise InternalError("import is missing nodes in network set")

        return import_order
